from __future__ import annotations

from tumfl.Token import Token

from .Statement import Statement


class Semicolon(Statement):
    """A semicolon"""

    def __init__(self, token: Token):
        super().__init__(token, "Semicolon")



import argparse
import logging
import time
from logging import error, info, warning
from pathlib import Path
from typing import Optional

from watchdog.events import (
    EVENT_TYPE_CLOSED,
    EVENT_TYPE_OPENED,
    FileSystemEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer

from tumfl import format, parse
from tumfl.AST import Assign, ASTNode, Name, String
from tumfl.basic_walker import NoneWalker
from tumfl.dependency_resolver import resolve_recursive
from tumfl.error import TumflError


class ReplaceWalker(NoneWalker):
    def __init__(self, replacements: dict[str, ASTNode], prefix: str = "$$"):
        super().__init__()
        self.replacements: dict[str, ASTNode] = replacements
        self.prefix: str = prefix

    def visit_String(self, node: String) -> None:
        if node.value.startswith(self.prefix):
            if node.value[len(self.prefix) :] in self.replacements:
                assert node.parent_class
                node.parent_class.replace_child(
                    node, self.replacements[node.value[len(self.prefix) :]]
                )
            else:
                warning(f"Found placeholder with no assigned value: {node.value}")


class ParameterGetter(NoneWalker):
    def __init__(self) -> None:
        super().__init__()
        self.parameters: dict[str, ASTNode] = {}

    def visit_Assign(self, node: Assign) -> None:
        if (
            len(node.targets) == 1
            and isinstance(node.targets[0], Name)
            and len(node.expressions) == 1
        ):
            name = node.targets[0]
            assert isinstance(name, Name)
            self.parameters[name.variable_name] = node.expressions[0]
        else:
            warning(f"Ignoring config parameter {node}")


def compile_file(filename: Path, replacer: ReplaceWalker) -> str:
    source_directory: Path = filename.parent
    result: ASTNode = resolve_recursive(
        filename, [source_directory], add_source_description=True
    )
    replacer.visit(result)
    return format(result)


def run(source: Path, destination: Optional[Path], replacer: ReplaceWalker) -> None:
    if not destination:
        destination = source.with_stem(source.stem + "_result")
    try:
        compiled = compile_file(source, replacer)
    except RuntimeError:
        return
    except TumflError as e:
        error(e)
        return
    with destination.open("W") as f:
        f.write(compiled)


class EventHandler(FileSystemEventHandler):
    def __init__(
        self, source: Path, destination: Optional[Path], replacer: ReplaceWalker
    ):
        self.source: Path = source
        self.destination: Optional[Path] = destination
        self.replacer: ReplaceWalker = replacer

    def execute(self) -> None:
        info("Files changed, recompiling")
        run(self.source, self.destination, self.replacer)

    def on_any_event(self, event: FileSystemEvent) -> None:
        if event.event_type not in (EVENT_TYPE_CLOSED, EVENT_TYPE_OPENED):
            self.execute()


def follow(source: Path, destination: Optional[Path], replacer: ReplaceWalker) -> None:
    run(source, destination, replacer)
    observer = Observer()
    event_handler = EventHandler(source, destination, replacer)
    observer.schedule(event_handler, str(source.parent), recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(100)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def parse_config(file: Path) -> dict[str, ASTNode]:
    with file.open() as f:
        chunk = parse(f.read())
    getter = ParameterGetter()
    getter.visit(chunk)
    return getter.parameters


def main() -> None:
    parser = argparse.ArgumentParser(description="Compile lua files", prog="tumfl")
    parser.add_argument("source_file", type=Path, help="Source file to compile")
    parser.add_argument("-d", "--destination", type=Path, help="Destination file")
    parser.add_argument(
        "-v",
        "--verbose",
        help="Be verbose",
        action="count",
        default=0,
    )
    parser.add_argument(
        "-f", "--follow", action="store_true", help="Follow file changes"
    )
    parser.add_argument(
        "-c",
        "--config-file",
        type=Path,
        help="Replace placeholders using config file."
        " Config file is a normal lua file, with top level assignments of `name = value`, not limited to strings."
        " To use an replacement, just use a string int the target file like"
        ' `"$$name"` (if `$$` is your prefix, and `name` the name to look up).',
    )
    parser.add_argument(
        "--config-prefix",
        default="$$",
        help="Prefix for names to be replaced by config values (default: $$)",
    )
    args = parser.parse_args()
    loglevel = logging.WARN
    if args.verbose == 1:
        loglevel = logging.INFO
    elif args.verbose > 1:
        loglevel = logging.DEBUG
    logging.basicConfig(
        format="%(asctime)s|%(levelname)s| %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        level=loglevel,
    )
    _replacer = ReplaceWalker({}, prefix=args.conf_prefix)
    if args.config_file:
        _replacer.replacements = parse_config(args.config_file)
    if args.follow:
        follow(args.source_file, args.destination, _replacer)
    else:
        run(args.source_file, args.destination, _replacer)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Optional

from tumfl.AST.Expression.Expression import Expression
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class Chunk(Block):
    """Same as block, but individually executable."""

    def __init__(
        self,
        token: Token,
        statements: list[Statement],
        returns: Optional[list[Expression]],
    ):
        super().__init__(token, statements, returns)
        self.name = "Chunk"



from __future__ import annotations

from typing import Optional

from tumfl.AST.Expression.Expression import Expression
from tumfl.AST.Expression.Name import Name
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class NumericFor(Statement):
    """The numeric for, i.e. for i=1,10 do"""

    def __init__(
        self,
        token: Token,
        variable_name: Name,
        start: Expression,
        stop: Expression,
        step: Optional[Expression],
        body: Block,
    ):
        super().__init__(token, "NumericFor")
        self.variable_name: Name = variable_name
        self.start: Expression = start
        self.stop: Expression = stop
        self.step: Optional[Expression] = step
        self.body: Block = body



from __future__ import annotations

from enum import Enum

from tumfl.Token import Token

from .Expression import Expression


class BinaryOperand(Enum):
    # pylint: disable=duplicate-code
    PLUS = "+"
    MINUS = "-"
    MULT = "*"
    DIVIDE = "/"
    INTEGER_DIVISION = "//"
    MODULO = "%"
    EXPONENT = "^"
    BIT_AND = "&"
    BIT_OR = "|"
    BIT_XOR = "~"
    BIT_SHIFT_LEFT = "<<"
    BIT_SHIFT_RIGHT = ">>"
    CONCAT = ".."
    EQUALS = "=="
    NOT_EQUALS = "~="
    LESS_EQUALS = "<="
    GREATER_EQUALS = ">="
    LESS_THAN = "<"
    GREATER_THAN = ">"
    AND = "and"
    OR = "or"

    @property
    def __lvl2(self) -> tuple[BinaryOperand, ...]:
        return (
            BinaryOperand.EQUALS,
            BinaryOperand.NOT_EQUALS,
            BinaryOperand.LESS_EQUALS,
            BinaryOperand.GREATER_EQUALS,
            BinaryOperand.LESS_THAN,
            BinaryOperand.GREATER_THAN,
        )

    @property
    def __bitwise_operands(self) -> tuple[BinaryOperand, ...]:
        return (
            BinaryOperand.BIT_XOR,
            BinaryOperand.BIT_AND,
            BinaryOperand.BIT_SHIFT_LEFT,
            BinaryOperand.BIT_SHIFT_RIGHT,
            BinaryOperand.CONCAT,
            BinaryOperand.PLUS,
            BinaryOperand.MINUS,
        )

    @property
    def __lvl8(self) -> tuple[BinaryOperand, ...]:
        return (
            BinaryOperand.MULT,
            BinaryOperand.DIVIDE,
            BinaryOperand.INTEGER_DIVISION,
            BinaryOperand.MODULO,
        )

    def get_precedence(self) -> int:
        if self == BinaryOperand.OR:
            return 0
        if self == BinaryOperand.AND:
            return 1
        if self in self.__lvl2:
            return 2
        if self == BinaryOperand.BIT_OR:
            return 3
        if self == BinaryOperand.BIT_XOR:
            return 4
        if self == BinaryOperand.BIT_AND:
            return 5
        if self in (BinaryOperand.BIT_SHIFT_RIGHT, BinaryOperand.BIT_SHIFT_LEFT):
            return 6
        if self == BinaryOperand.CONCAT:
            return 7
        if self in (BinaryOperand.PLUS, BinaryOperand.MINUS):
            return 8
        if self in self.__lvl8:
            return 9
        # precedence 10 is unop
        if self == BinaryOperand.EXPONENT:
            return 11
        assert False, f"Unknown Binary Operand {self}"

    def get_optional_brackets(self) -> tuple[BinaryOperand, ...]:
        if self == BinaryOperand.OR:
            return (BinaryOperand.AND,)
        if self == BinaryOperand.AND:
            return ()
        if self in self.__lvl2:
            return self.__lvl2
        if self == BinaryOperand.BIT_OR:
            return self.__bitwise_operands
        if self == BinaryOperand.BIT_XOR:
            return self.__bitwise_operands
        if self == BinaryOperand.BIT_AND:
            return self.__bitwise_operands
        if self in (BinaryOperand.BIT_SHIFT_RIGHT, BinaryOperand.BIT_SHIFT_LEFT):
            return self.__bitwise_operands
        if self == BinaryOperand.CONCAT:
            return BinaryOperand.PLUS, BinaryOperand.MINUS
        if self in (BinaryOperand.PLUS, BinaryOperand.MINUS):
            return self.__lvl8
        if self in self.__lvl8:
            return (BinaryOperand.EXPONENT,)
        # precedence 10 is unop
        if self == BinaryOperand.EXPONENT:
            return ()
        assert False, f"Unknown Binary Operand {self}"

    def non_commutative(self) -> bool:
        if self in (BinaryOperand.MINUS, BinaryOperand.DIVIDE, BinaryOperand.EXPONENT):
            return True
        return False

    def __repr__(self) -> str:
        return f"BinaryOperand.{self.name}"


class BinOp(Expression):
    """Binary Operation, like a + b, 1 and 2 or 4 == b"""

    def __init__(
        self, token: Token, op: BinaryOperand, left: Expression, right: Expression
    ) -> None:
        super().__init__(token, "BinOp")
        self.op: BinaryOperand = op
        self.left: Expression = left
        self.right: Expression = right

    @staticmethod
    def from_token(token: Token, left: Expression, right: Expression) -> BinOp:
        binop: BinaryOperand = BinaryOperand[token.type.name]
        assert binop
        return BinOp(token, binop, left, right)



from __future__ import annotations

from pathlib import Path
from typing import Optional, Union

from .AST import (
    ASTNode,
    Chunk,
    ExpFunctionCall,
    ExpFunctionDefinition,
    FunctionCall,
    Name,
    Semicolon,
    String,
)
from .basic_walker import NoneWalker
from .error import InvalidDependencyError
from .parser import parse
from .Token import Token, TokenType


def _parse_file(path: Path) -> Chunk:
    with path.open() as f:
        chunk: str = f.read()
    ast = parse(chunk)
    ast.parent(None, path)
    return ast


class ResolveDependencies(NoneWalker):
    def __init__(self, search_path: list[Path], add_source_description: bool = False):
        self.search_path: list[Path] = search_path
        self.found: dict[Path, Optional[list[Name]]] = {}
        self.add_source_description: bool = add_source_description

    @staticmethod
    def _str_to_name(name: str) -> Name:
        return Name.from_token(Token(TokenType.NAME, name, 0, 0))

    def _find_file_in_path(self, name: str, start_path: Path) -> Optional[Path]:
        parts: list[str] = name.split(".")
        if not parts[0]:
            return None
        filename: Path = Path(parts.pop(0))
        for file in parts:
            filename /= file
        for path in (start_path, *self.search_path):
            for suffix in ("", ".tl", ".lua"):
                current_path = path.joinpath(filename.with_suffix(suffix))
                if current_path.is_file():
                    return current_path
        return None

    def _get_block_dependency_path(
        self, name: str, start_path: Path, token: Token, deduplicate: bool = True
    ) -> Optional[Path]:
        path: Optional[Path] = self._find_file_in_path(name, start_path)
        if not path:
            raise InvalidDependencyError(
                f"Could not find dependency with name {name}", token
            )
        if deduplicate and path in self.found:
            return None
        self.found[path] = None
        return path

    def __get_ast(
        self, node: Union[FunctionCall, ExpFunctionCall], deduplicate: bool = True
    ) -> Optional[Chunk | Semicolon]:
        if isinstance(node.function, Name) and node.function.variable_name == "require":
            if len(node.arguments) == 1 and isinstance(
                name := node.arguments[0], String
            ):
                assert isinstance(name, String)
                assert node.file_name
                dependency_path: Optional[Path] = self._get_block_dependency_path(
                    name.value, node.file_name.parent, node.token, deduplicate
                )
                assert node.parent_class
                ast: Union[Chunk | Semicolon]
                if dependency_path:
                    ast = _parse_file(dependency_path)
                    if self.add_source_description:
                        ast.comment.insert(0, f"Sourced from {dependency_path}")
                else:
                    ast = Semicolon(node.token)
                    ast.parent(node.parent_class, node.file_name)
                return ast
            raise InvalidDependencyError(
                f"Wrong require() arguments. Expected single string, got {node.arguments}",
                node.token,
            )
        return None

    def visit_FunctionCall(self, node: FunctionCall) -> None:
        if ast := self.__get_ast(node):
            assert node.parent_class
            node.parent_class.replace_child(node, ast)
            ast.parent_class = node.parent_class
            self.visit(ast)
        else:
            super().visit_FunctionCall(node)

    def visit_ExpFunctionCall(self, node: ExpFunctionCall) -> None:
        # if the body is expecting a return value, never deduplicate
        if ast := self.__get_ast(node, deduplicate=False):
            # since deduplication is off, ast will never be a Semicolon
            assert isinstance(ast, Chunk)
            function = ExpFunctionDefinition(node.token, [], ast)
            node.function = function
            ast.parent_class = function
            ast.file_name = node.file_name
            function.parent_class = node
            self.visit(function)
        else:
            super().visit_ExpFunctionCall(node)


def resolve_recursive(
    path: Path, search_path: list[Path], add_source_description: bool = False
) -> ASTNode:
    ast = _parse_file(path)
    resolver = ResolveDependencies(search_path, add_source_description)
    resolver.visit(ast)
    return ast



from __future__ import annotations

from tumfl.AST.BaseFunctionDefinition import BaseFunctionDefinition
from tumfl.AST.Expression.Name import Name
from tumfl.AST.Expression.Vararg import Vararg
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class LocalFunctionDefinition(Statement):
    """Definition for a local function"""

    def __init__(
        self,
        token: Token,
        function_name: Name,
        parameters: list[Name | Vararg],
        body: Block,
    ):
        super().__init__(token, "LocalFunctionDefinition")
        self.function_name: Name = function_name
        self.parameters: list[Name | Vararg] = parameters
        self.body: Block = body

    @staticmethod
    def from_base_definition(
        base: BaseFunctionDefinition, function_name: Name
    ) -> LocalFunctionDefinition:
        return LocalFunctionDefinition(
            base.token, function_name, base.parameters, base.body
        )



from __future__ import annotations

from typing import Optional

from tumfl.Token import Token, TokenType

from .Expression import Expression


class Number(Expression):
    """A number, still preserved in the original representation, like 1.2, 3 or 0x123f"""

    def __init__(
        self,
        token: Token,
        is_hex: bool = False,
        integer_part: Optional[str] = None,
        fractional_part: Optional[str] = None,
        exponent: Optional[str] = None,
        float_offset: Optional[str] = None,
    ):
        super().__init__(token, "Number")
        self.is_hex: bool = is_hex
        self.integer_part: Optional[str] = integer_part
        self.fractional_part: Optional[str] = fractional_part
        self.exponent: Optional[str] = exponent
        self.float_offset: Optional[str] = float_offset

    @staticmethod
    def from_token(token: Token) -> Number:
        assert token.type == TokenType.NUMBER
        value = token.value
        assert isinstance(value, tuple)
        return Number(
            token,
            is_hex=value[0],
            integer_part=value[1],
            fractional_part=value[2],
            exponent=value[3],
            float_offset=value[4],
        )

    def to_int(self) -> Optional[int]:
        if self.exponent or self.fractional_part or self.float_offset:
            # is a float
            return None
        assert self.integer_part
        return int(self.integer_part, 16) if self.is_hex else int(self.integer_part)

    def to_float(self) -> float:
        str_repr: str = str(self)
        if self.is_hex:
            return float.fromhex(str_repr)
        return float(str_repr)

    def __str__(self) -> str:
        str_repr: str = ""
        if self.is_hex:
            str_repr += "0x"
        if self.integer_part:
            str_repr += self.integer_part
        else:
            str_repr += "1" if self.is_hex else "0"
        if self.fractional_part:
            str_repr += f".{self.fractional_part}"
        if self.exponent:
            str_repr += f"e{self.exponent}"
        if self.float_offset:
            str_repr += f"p{self.float_offset}"
        return str_repr



from __future__ import annotations

from tumfl.Token import Token, TokenType

from .Variable import Variable


class Name(Variable):
    """The name of the variable, like a, b or very_long_variable_name, NOT a.b"""

    def __init__(self, token: Token, name: str):
        super().__init__(token, "Name")
        self.variable_name: str = name

    @staticmethod
    def from_token(token: Token) -> Name:
        assert token.type == TokenType.NAME
        assert isinstance(token.value, str)
        return Name(token, token.value)

    def __str__(self) -> str:
        return self.variable_name



from __future__ import annotations

import string
from enum import Enum
from typing import Iterator, Literal, Optional, Sequence, Type

from .AST import *
from .basic_walker import BasicWalker


class FormattingStyle:
    # pylint: disable=too-few-public-methods

    # Statement separator
    STATEMENT_SEPARATOR: str = "\n"
    # Indentation character
    INDENTATION: str = "\t"
    # Argument separator
    ARGUMENT_SEPARATOR: str = ", "
    # Include comments or omit them
    INCLUDE_COMMENTS: bool = True
    # Separator after --
    COMMENT_SEP: str = " "
    # Use single quotes if this saves on string escapes
    USE_SINGLE_QUOTE: bool = False
    # Use direct table and string calls
    USE_CALL_SHORTHAND: bool = False
    # Remove all characters that are not strictly required
    REMOVE_UNNECESSARY_CHARS: bool = False
    # Add brackets to arithmetic expressions no matter the use
    ADD_ALL_BRACKETS: bool = False
    # Add brackets to arithmetic expressions if the order is potentially confusing
    ADD_CLOSE_BRACKETS: bool = True
    # Add a space around `=` in table constructors
    SPACE_IN_TABLE: bool = True
    # highest number of newline characters before using a multiline string
    NEWLINE_LIMIT: int = 4
    # How big lines may be. Note: there is no guarantee that this will be honored. Use 0 for no limit
    LINE_WIDTH: int = 120
    # Number of lines in a block to add spacers before and after the block. Use 0 to disable
    BLOCK_SPACER: int = 5
    # Keep semicolons
    KEEP_SEMICOLON: bool = False


class MinifiedStyle(FormattingStyle):
    # pylint: disable=too-few-public-methods

    STATEMENT_SEPARATOR = ";"
    INDENTATION = ""
    ARGUMENT_SEPARATOR = ","
    INCLUDE_COMMENTS = False
    COMMENT_SEP = ""
    USE_SINGLE_QUOTE = True
    USE_CALL_SHORTHAND = True
    REMOVE_UNNECESSARY_CHARS = True
    NEWLINE_LIMIT = 1
    LINE_WIDTH = 0
    BLOCK_SPACER = 0


class Separators(Enum):
    Statement = "stmt"
    Newline = "newline"
    Argument = "arg"
    Space = " "
    Dot = "."
    Indent = "indent"
    DeIndent = "de-indent"
    Block = "block"

    def join(self, to_join: Iterator[Retype]) -> Retype:
        join_list: list[Retype] = list(to_join)
        for i in range(len(join_list) - 1, 0, -1):
            join_list[i:i] = [[self]]
        return [elem for part in join_list for elem in part]

    def __repr__(self) -> str:
        return f"Separators.{self.name}"


Retype = list[str | Separators]
ESCAPE_CHARACTERS: dict[str, str] = {
    "\a": "a",
    "\b": "b",
    "\f": "f",
    "\n": "n",
    "\r": "r",
    "\t": "t",
    "\v": "v",
}


class Formatter(BasicWalker[Retype]):
    # pylint: disable=too-many-public-methods

    def __init__(self, style: Type[FormattingStyle]):
        self.s: Type[FormattingStyle] = style

    def _format_args(self, arguments: Sequence[ASTNode]) -> Retype:
        return Separators.Argument.join(self.visit(arg) for arg in arguments)

    @staticmethod
    def _find_level(to_check: str) -> int:
        level: int = 0
        if to_check.endswith("]"):
            level = 1
        while True:
            if (
                f"[{'=' * level}[" not in to_check
                and f"]{'=' * level}]" not in to_check
            ):
                break
            level += 1
        return level

    def _format_comment(self, comment: str, force_long: bool = False) -> Retype:
        comment = comment.strip()
        if force_long or "\n" in comment:
            level: int = self._find_level(comment)
            return [f"--[{'=' * level}[{comment}]{'=' * level}]", Separators.Statement]
        return [f"--{self.s.COMMENT_SEP}{comment}", Separators.Newline]

    def _format_function_args(self, args: Sequence[Expression]) -> Retype:
        if self.s.USE_CALL_SHORTHAND and len(args) == 1:
            if isinstance(args[0], String):
                return self.visit(args[0])
            if isinstance(args[0], Table):
                return self.visit(args[0])
        return ["(", *self._format_args(args), ")"]

    def _format_var(self, var: Expression) -> Retype:
        if isinstance(
            var, (Name, Index, NamedIndex, ExpFunctionCall, ExpMethodInvocation)
        ):
            return self.visit(var)
        return ["(", *self.visit(var), ")"]

    def visit_Assign(self, node: Assign) -> Retype:
        targets: list[Retype] = [self._format_var(var) for var in node.targets]
        return [
            *Separators.Argument.join(iter(targets)),
            Separators.Space,
            "=",
            Separators.Space,
            *self._format_args(node.expressions),
        ]

    def visit_Block(self, node: Block) -> Retype:
        result: Retype = ["do", Separators.Block, Separators.Indent]
        for statement in node.statements:
            if self.s.INCLUDE_COMMENTS and statement.comment:
                for comment in statement.comment:
                    result += self._format_comment(comment)
            result += self.visit(statement) + [Separators.Statement]
        if node.returns is not None:
            opt_space: tuple[Separators, ...] = (
                (Separators.Space,) if node.returns else ()
            )
            returns: Retype = self._format_args(node.returns)
            result += ["return", *opt_space, *returns, Separators.Statement]
        result += [Separators.DeIndent, "end"]
        return result

    def visit_Boolean(self, node: Boolean) -> Retype:
        return [str(node.value).lower()]

    def visit_Break(self, _node: Break) -> Retype:
        return ["break"]

    def visit_Chunk(self, node: Chunk) -> Retype:
        result: Retype = self.visit_Block(node)
        return result[3:-3]

    def visit_Goto(self, node: Goto) -> Retype:
        return ["goto", Separators.Space, *self.visit(node.label_name)]

    def visit_If(self, node: If) -> Retype:
        spaced_test: Retype = [
            Separators.Space,
            *self.visit(node.test),
            Separators.Space,
        ]
        result: Retype = ["if", *spaced_test, "then", Separators.Block]
        result += self.visit(node.true)[2:-1]
        current_if: If = node
        while isinstance(current_if.false, If):
            current_if = current_if.false
            spaced_test = [
                Separators.Space,
                *self.visit(current_if.test),
                Separators.Space,
            ]
            result += ["elseif", *spaced_test, "then", Separators.Block]
            result += self.visit(current_if.true)[2:-1]
        if current_if.false:
            result += ["else", Separators.Block]
            result += self.visit(current_if.false)[2:-1]
        result += ["end"]
        return result

    def visit_Index(self, node: Index) -> Retype:
        return [*self._format_var(node.lhs), "[", *self.visit(node.variable_name), "]"]

    def visit_Label(self, node: Label) -> Retype:
        return ["::", *self.visit(node.label_name), "::"]

    def visit_Name(self, node: Name) -> Retype:
        return [node.variable_name]

    def visit_Nil(self, _node: Nil) -> Retype:
        return ["nil"]

    def visit_Number(self, node: Number) -> Retype:
        return [str(node)]

    def visit_Repeat(self, node: Repeat) -> Retype:
        return [
            "repeat",
            Separators.Block,
            *self.visit(node.body)[2:-1],
            "until",
            Separators.Space,
            *self.visit(node.condition),
        ]

    def visit_Semicolon(self, _node: Semicolon) -> Retype:
        return [";"] if self.s.KEEP_SEMICOLON else []

    def visit_String(self, node: String) -> Retype:
        escaped: str = ""
        quote: str = '"'
        contains_unprintable: bool = False
        single_quote_count: int = 0
        double_quote_count: int = 0
        newline_count: int = 0
        for char in node.value:
            if char == "\n":
                newline_count += 1
            elif char == "'":
                single_quote_count += 1
            elif char == '"':
                double_quote_count += 1
            elif not 32 <= ord(char) < 127:
                contains_unprintable = True
        # do not use multiline strings if the number of newlines is too low
        if newline_count > self.s.NEWLINE_LIMIT and not contains_unprintable:
            level: int = self._find_level(node.value)
            return [f"[{'=' * level}[{node.value}]{'=' * level}]"]
        if self.s.USE_SINGLE_QUOTE and single_quote_count < double_quote_count:
            quote = "'"
        for char in node.value:
            if char in (quote, "\\"):
                escaped += f"\\{char}"
            elif 32 <= ord(char) < 127:
                escaped += char
            elif char in ESCAPE_CHARACTERS:
                escaped += f"\\{ESCAPE_CHARACTERS[char]}"
            else:
                escaped += f"\\x{ord(char):02x}"
        return [quote + escaped + quote]

    def visit_Table(self, node: Table) -> Retype:
        return ["{", *self._format_args(node.fields), "}"]

    def visit_Vararg(self, _node: Vararg) -> Retype:
        return ["..."]

    def visit_While(self, node: While) -> Retype:
        return [
            "while",
            Separators.Space,
            *self.visit(node.condition),
            Separators.Space,
            *self.visit(node.body),
        ]

    def _need_brackets(
        self, own_node: BinOp, other_node: Expression, care_unop: bool
    ) -> bool:
        own_precedence: int = own_node.op.get_precedence()
        return (
            self.s.ADD_ALL_BRACKETS
            or isinstance(other_node, BinOp)
            and (
                own_precedence > other_node.op.get_precedence()
                or self.s.ADD_CLOSE_BRACKETS
                and other_node.op in own_node.op.get_optional_brackets()
                or other_node.op.non_commutative()
                and own_precedence == other_node.op.get_precedence()
            )
            or (care_unop or self.s.ADD_CLOSE_BRACKETS)
            and isinstance(other_node, UnOp)
            and (
                own_precedence > 10 or self.s.ADD_CLOSE_BRACKETS and own_precedence >= 9
            )
        )

    def visit_BinOp(self, node: BinOp) -> Retype:
        result: Retype = []
        if self._need_brackets(node, node.left, True):
            result += ["(", *self.visit(node.left), ")"]
        else:
            result += self.visit(node.left)
        result += [Separators.Space, node.op.value, Separators.Space]
        if self._need_brackets(node, node.right, False):
            result += ["(", *self.visit(node.right), ")"]
        else:
            result += self.visit(node.right)
        return result

    def visit_FunctionCall(self, node: FunctionCall) -> Retype:
        return self._format_var(node.function) + self._format_function_args(
            node.arguments
        )

    def visit_FunctionDefinition(self, node: FunctionDefinition) -> Retype:
        full_name: Retype = Separators.Dot.join(self.visit(name) for name in node.names)
        if node.method_name:
            full_name += [":", *self.visit(node.method_name)]
        return [
            Separators.Newline,
            "function",
            Separators.Space,
            *full_name,
            "(",
            *self._format_args(node.parameters),
            ")",
            *self.visit(node.body)[1:],
            Separators.Block,
            Separators.Newline,
        ]

    def visit_IterativeFor(self, node: IterativeFor) -> Retype:
        return [
            "for",
            Separators.Space,
            *self._format_args(node.namelist),
            Separators.Space,
            "in",
            Separators.Space,
            *self._format_args(node.explist),
            Separators.Space,
            *self.visit(node.body),
        ]

    def visit_LocalAssign(self, node: LocalAssign) -> Retype:
        result: Retype = ["local", Separators.Space]
        names: list[Retype] = []
        for name in node.variable_names:
            if name.attribute:
                names.append(
                    [str(name.name), Separators.Space, "<", str(name.attribute), ">"]
                )
            else:
                names.append([str(name.name)])
        result += Separators.Argument.join(iter(names))
        if node.expressions:
            result += [Separators.Space, "=", Separators.Space]
            result += self._format_args(node.expressions)
        return result

    def visit_MethodInvocation(self, node: MethodInvocation) -> Retype:
        return [
            *self._format_var(node.function),
            ":",
            *self.visit(node.method),
            *self._format_function_args(node.arguments),
        ]

    def visit_NamedIndex(self, node: NamedIndex) -> Retype:
        return [
            *self._format_var(node.lhs),
            Separators.Dot,
            *self.visit(node.variable_name),
        ]

    def visit_NumericFor(self, node: NumericFor) -> Retype:
        spaced_name: Retype = [
            Separators.Space,
            *self.visit(node.variable_name),
            Separators.Space,
        ]
        result: Retype = ["for", *spaced_name, "=", Separators.Space]
        result += self.visit(node.start)
        result += [Separators.Argument, *self.visit(node.stop)]
        if node.step:
            result += [Separators.Argument, *self.visit(node.step)]
        result.append(Separators.Space)
        result += self.visit(node.body)
        return result

    def visit_UnOp(self, node: UnOp) -> Retype:
        result: Retype = [node.op.value]
        if (
            self.s.ADD_ALL_BRACKETS
            or isinstance(node.right, BinOp)
            and node.right.op != BinaryOperand.EXPONENT
        ):
            result += ["(", *self.visit(node.right), ")"]
        elif isinstance(node.right, UnOp) or node.op == UnaryOperand.NOT:
            result += [Separators.Space, *self.visit(node.right)]
        else:
            result += self.visit(node.right)
        return result

    def visit_ExpFunctionCall(self, node: ExpFunctionCall) -> Retype:
        return self._format_var(node.function) + self._format_function_args(
            node.arguments
        )

    def visit_ExpFunctionDefinition(self, node: ExpFunctionDefinition) -> Retype:
        return [
            "function",
            Separators.Space,
            "(",
            *self._format_args(node.parameters),
            ")",
            *self.visit(node.body)[1:],
        ]

    def visit_ExpMethodInvocation(self, node: ExpMethodInvocation) -> Retype:
        return [
            *self._format_var(node.function),
            ":",
            *self.visit(node.method),
            *self._format_function_args(node.arguments),
        ]

    def visit_LocalFunctionDefinition(self, node: LocalFunctionDefinition) -> Retype:
        return [
            Separators.Newline,
            "local",
            Separators.Space,
            "function",
            Separators.Space,
            *self.visit(node.function_name),
            "(",
            *self._format_args(node.parameters),
            ")",
            *self.visit(node.body)[1:],
            Separators.Statement,
            Separators.Newline,
        ]

    def visit_ExplicitTableField(self, node: ExplicitTableField) -> Retype:
        return [
            "[",
            *self.visit(node.at),
            "]",
            Separators.Space,
            "=",
            Separators.Space,
            *self.visit(node.value),
        ]

    def visit_NamedTableField(self, node: NamedTableField) -> Retype:
        return [
            *self.visit(node.field_name),
            Separators.Space,
            "=",
            Separators.Space,
            *self.visit(node.value),
        ]

    def visit_NumberedTableField(self, node: NumberedTableField) -> Retype:
        return self.visit(node.value)


def sep_required(first: str, second: str) -> bool:
    """
    Returns, whether a separator is strictly necessary between 2 tokens

    :param first: last character of first token
    :param second: first character of second token
    :return: if the token is required
    """
    return (
        first in string.ascii_letters + string.digits
        and second in string.ascii_letters + string.digits
        or first == "-"
        and second == "-"
    )


def search_token(
    start_idx: int, direction: Literal[1, -1], tokens: Retype
) -> str | Separators:
    """
    Search for a non indent/de-indent token.

    :param start_idx: the first index to search
    :param direction: in what direction to search, 1 for forwards and -1 for backwards
    :param tokens: the token stream
    :return: the closest non-indent token
    """
    while isinstance(token := tokens[start_idx], Separators) and token in (
        Separators.Indent,
        Separators.DeIndent,
    ):
        start_idx += direction
    return token


def remove_separators(token_stream: Retype) -> None:
    """
    Remove all seperator tokens that are not strictly necessary.

    :param token_stream: The token stream to filter. Removal happens in-place
    :return: None
    """
    # Add a dummy token so that there is definitely a valid final token to look for
    token_stream.append("/")
    for i in range(len(token_stream) - 2, 0, -1):
        if isinstance(own_token := token_stream[i], Separators):
            if own_token not in (
                Separators.Space,
                Separators.Statement,
                Separators.Block,
            ):
                continue
            previous_token = search_token(i - 1, -1, token_stream)
            next_token = search_token(i + 1, 1, token_stream)
            if (
                isinstance(previous_token, Separators)
                or isinstance(next_token, Separators)
                or not sep_required(previous_token[-1], next_token[0])
            ):
                token_stream.pop(i)
    # Remove the dummy token again
    token_stream.pop()


def __get_indentation_width(style: Type[FormattingStyle]) -> int:
    return len(style.INDENTATION.replace("\t", "    "))


def __estimate_width(
    indentation: int,
    stream: Retype,
    style: Type[FormattingStyle],
    with_indentation: bool = True,
) -> Optional[int]:
    """
    Estimate the width of a stream

    :param indentation: indentation of the stream
    :param stream: The stream to estimate
    :param style: Formatting style to apply
    :return: None if multiline, estimated width otherwise
    """
    size: int = 0
    if with_indentation:
        size = indentation * __get_indentation_width(style)
    for token in stream:
        match token:
            case Separators.Space | Separators.Dot:
                size += 1
            case Separators.Argument:
                size += len(style.ARGUMENT_SEPARATOR)
            case Separators.Newline | Separators.Statement | Separators.Block:
                return None
            case Separators.Indent | Separators.DeIndent:
                pass
            case literal_string:
                if "\n" in literal_string:
                    return None
                size += len(literal_string)
    return size


def __get_newline_pos(input_string: str, max_pos: int) -> int:
    if len(input_string) < max_pos:
        return len(input_string)
    current_pos: int = max_pos
    while current_pos > 1:
        is_end: bool = current_pos == len(input_string)
        is_allowed: bool = is_end or not input_string[current_pos].isspace()
        is_word_break: bool = not input_string[current_pos - 1].isalnum()
        if is_allowed and is_word_break:
            return current_pos
        current_pos -= 1
    current_pos = max_pos
    while current_pos < len(input_string):
        if not input_string[current_pos].isspace():
            return current_pos
        current_pos += 1
    return len(input_string)


def _string_ident(
    input_string: str, indentation: int, style: Type[FormattingStyle]
) -> Retype:
    start: str = input_string[0]
    assert start in "'\""
    assert input_string[-1] == start
    width = __estimate_width(indentation, [input_string], style)
    if width and width <= style.LINE_WIDTH:
        return [input_string]
    raw_indentation: int = indentation * __get_indentation_width(style) + 2
    result: Retype = []
    while input_string:
        pos = __get_newline_pos(input_string, style.LINE_WIDTH - raw_indentation)
        if pos >= len(input_string) - 2:
            pos = len(input_string)
        result.append(input_string[:pos] + "\\z")
        result.append(Separators.Newline)
        input_string = input_string[pos:]
    # remove last newline
    result.pop()
    last_string = result[-1]
    assert isinstance(last_string, str)
    # remove last \z
    result[-1] = last_string[:-2]
    return result


MATCHING_BRACKETS: dict[str, str] = {"}": "{", "]": "[", ")": "("}


def __inner_indent(
    token_stream: Retype, index: int, indentation: int, style: Type[FormattingStyle]
) -> tuple[int, Retype]:
    opening: str | Separators = token_stream[index]
    assert opening in MATCHING_BRACKETS
    closing: str = MATCHING_BRACKETS[opening]  # type: ignore
    # lua functions don't allow trailing argument separators
    use_trailing: bool = opening != ")"
    components: list[Retype] = []
    current_component: Retype = []
    index -= 1
    while True:
        current_char = token_stream[index]
        if current_char == closing:
            break
        if current_char in MATCHING_BRACKETS:
            index, content = __inner_indent(token_stream, index, indentation + 1, style)
            current_component[0:0] = content
        elif current_char == Separators.Argument:
            components.insert(0, current_component)
            current_component = []
        elif isinstance(current_char, str) and (
            current_char.startswith('"') or current_char.startswith("'")
        ):
            current_component[0:0] = _string_ident(current_char, indentation + 1, style)
        else:
            current_component.insert(0, current_char)
        index -= 1
    if current_component:
        components.insert(0, current_component)
    width: int = 0
    for component in components:
        current_width: Optional[int] = __estimate_width(
            indentation, component, style, with_indentation=False
        )
        if current_width is None:
            width = style.LINE_WIDTH + 1
            break
        width += current_width + len(style.ARGUMENT_SEPARATOR)
    if (
        width + 2 + indentation * __get_indentation_width(style) <= style.LINE_WIDTH
        or len(components) <= 1
    ):
        return index, [closing, *Separators.Argument.join(iter(components)), opening]
    result: Retype = [closing, Separators.Indent, Separators.Newline]
    for component in components:
        result.extend(component)
        result.append(Separators.Argument)
        result.append(Separators.Newline)
    if not use_trailing:
        result.pop(len(result) - 2)
    result.append(Separators.DeIndent)
    result.append(opening)
    return index, result


def indent_brackets(token_stream: Retype, style: Type[FormattingStyle]) -> None:
    indentation: int = 0
    index = len(token_stream) - 1
    while index >= 0:
        token = token_stream[index]
        # iterating backwards
        if token == Separators.DeIndent:
            indentation += 1
        elif token == Separators.Indent:
            indentation -= 1
        elif token in MATCHING_BRACKETS:
            start_index = index
            index, content = __inner_indent(token_stream, index, indentation, style)
            token_stream[index : start_index + 1] = content
        elif isinstance(token, str) and (
            token.startswith('"') or token.startswith("'")
        ):
            token_stream[index : index + 1] = _string_ident(token, indentation, style)
        index -= 1


def __inner_add_spacing(
    token_stream: Retype, index: int, to_add: set[int], style: Type[FormattingStyle]
) -> tuple[int, int]:
    last_statement_idx: Optional[int] = None
    total_line_breaks: int = 0
    current_line_breaks: int = 0
    while index < len(token_stream):
        token = token_stream[index]
        if token == Separators.Indent:
            index, newlines = __inner_add_spacing(
                token_stream, index + 1, to_add, style
            )
            current_line_breaks += newlines
        elif token == Separators.DeIndent:
            return index, total_line_breaks + current_line_breaks
        elif token == Separators.Newline:
            current_line_breaks += 1
        elif isinstance(token, str):
            current_line_breaks += token.count("\n")
        elif token == Separators.Statement:
            total_line_breaks += current_line_breaks + 1
            if (
                current_line_breaks > style.BLOCK_SPACER
                and last_statement_idx is not None
            ):
                to_add.add(last_statement_idx + 1)
                to_add.add(index + 1)
            last_statement_idx = index
            current_line_breaks = 0
        index += 1
    return index, total_line_breaks + current_line_breaks


def add_spacing(token_stream: Retype, style: Type[FormattingStyle]) -> None:
    to_add: set[int] = set()
    __inner_add_spacing(token_stream, 0, to_add, style)
    for i in sorted(to_add, reverse=True):
        token_stream.insert(i, Separators.Newline)


def resolve_tokens(token_stream: Retype, style: Type[FormattingStyle]) -> None:
    """
    Resolves all tokens except for Indent and DeIndent to strings.

    :param token_stream: the token stream to resolve. Resolving happens in-place
    :param style: The style object
    :return: None
    """
    newline: str = (
        style.STATEMENT_SEPARATOR if "\n" in style.STATEMENT_SEPARATOR else "\n"
    )
    for i, token in enumerate(token_stream):
        if isinstance(token, Separators):
            if token == Separators.Space:
                token_stream[i] = " "
            elif token == Separators.Dot:
                token_stream[i] = "."
            elif token in (Separators.Statement, Separators.Block):
                token_stream[i] = style.STATEMENT_SEPARATOR
            elif token == Separators.Argument:
                if token_stream[i + 1] == Separators.Newline:
                    token_stream[i] = style.ARGUMENT_SEPARATOR.rstrip()
                else:
                    token_stream[i] = style.ARGUMENT_SEPARATOR
            elif token == Separators.Newline:
                token_stream[i] = newline
                i += 1
                while i < len(token_stream) and token_stream[i] == Separators.Newline:
                    token_stream[i] = ""


def indent(token_stream: Retype, indentation_str: str) -> None:
    """
    Indent the token stream, adding indentation_str once per indentation level in front of existing tokens.

    :param token_stream: the token stream to indent. Indentation happens in-place.
    :param indentation_str: the string to use for indenting
    :return: None
    """
    indentation_level: int = 0
    indentation_dirty: bool = False
    for i, token in enumerate(token_stream):
        if isinstance(token, Separators):
            assert token in (Separators.Indent, Separators.DeIndent)
            if token == Separators.Indent:
                indentation_level += 1
            else:
                indentation_level -= 1
        else:
            if indentation_dirty:
                token_stream[i] = indentation_level * indentation_str + token
                indentation_dirty = False
            if token.endswith("\n"):
                indentation_dirty = True
    assert indentation_level == 0


def join_tokens(token_stream: Retype) -> str:
    """
    Join all tokens into a string, ignoring all remaining Separators.

    :param token_stream: The token stream to join
    :return: The joined string
    """
    result: str = ""
    for token in token_stream:
        if isinstance(token, str):
            result += token
    return result


def __remove_orphaned_tokens(token_stream: Retype) -> None:
    for i in range(len(token_stream) - 1, -1, -1):
        token = token_stream[i]
        if token == "":
            token_stream.pop(i)
        elif token == Separators.Statement:
            if not isinstance(token_stream[i - 1], str):
                token_stream.pop(i)
            elif token_stream[i : i + 3] == [
                Separators.Statement,
                Separators.Newline,
                Separators.DeIndent,
            ]:
                token_stream.pop(i)


def format(ast: ASTNode, style: Optional[Type[FormattingStyle]] = None) -> str:
    """
    Format an ast tree according to the style class.

    :param ast: The AST Node to format, usually a chunk
    :param style: Optional style definition
    :return: Formatted lua
    """
    style = style or FormattingStyle
    formatter: Formatter = Formatter(style)
    token_stream: Retype = formatter.visit(ast)
    if style.REMOVE_UNNECESSARY_CHARS:
        remove_separators(token_stream)
    if style.LINE_WIDTH > 0:
        indent_brackets(token_stream, style)
    if style.BLOCK_SPACER > 0:
        add_spacing(token_stream, style)
    token_stream[0:0] = [f"--{style.COMMENT_SEP}tumfl", Separators.Newline]
    __remove_orphaned_tokens(token_stream)
    resolve_tokens(token_stream, style)
    indent(token_stream, style.INDENTATION)
    end = style.STATEMENT_SEPARATOR
    if style.REMOVE_UNNECESSARY_CHARS:
        end = ""
    formatted = join_tokens(token_stream)
    return "\n".join(line.rstrip() for line in formatted.split("\n")).strip() + end



from __future__ import annotations

from tumfl.AST.Expression.Name import Name
from tumfl.Token import Token

from .Statement import Statement


class Label(Statement):
    """A label to be used with goto"""

    def __init__(self, token: Token, label_name: Name):
        super().__init__(token, "Label")
        self.label_name: Name = label_name



import unittest

from tumfl.AST import Block, ExpFunctionDefinition, If
from tumfl.formatter import (
    Formatter,
    FormattingStyle,
    MinifiedStyle,
    Separators,
    _string_ident,
    format,
    indent,
    remove_separators,
    resolve_tokens,
    sep_required,
)
from tumfl.parser import Parser


class TestStyle(FormattingStyle):
    LINE_WIDTH = 7
    KEEP_SEMICOLON = True


class TestFormatter(unittest.TestCase):
    def setUp(self) -> None:
        self.normal: Formatter = Formatter(FormattingStyle)
        self.minified: Formatter = Formatter(MinifiedStyle)
        self.test: Formatter = Formatter(TestStyle)

    def test_format_comment(self):
        expected = ["-- comment", Separators.Newline]
        self.assertEqual(self.normal._format_comment("comment"), expected)
        expected = ["--comment", Separators.Newline]
        self.assertEqual(self.minified._format_comment("comment"), expected)
        expected = ["--[[ab\ncd]]", Separators.Statement]
        self.assertEqual(self.normal._format_comment("ab\ncd"), expected)
        expected = ["--[==[[[=[\na]==]", Separators.Statement]
        self.assertEqual(self.normal._format_comment("[[=[\na"), expected)

    def test_format_function_args(self):
        parser = Parser("'abc'{}1'ghi','jkl'")
        expression = parser._parse_exp()  # "abc"
        expected = ["(", *self.normal.visit(expression), ")"]
        self.assertEqual(self.normal._format_function_args([expression]), expected)
        expected = self.normal.visit(expression)
        self.assertEqual(self.minified._format_function_args([expression]), expected)
        expression = parser._parse_exp()  # {}
        expected = ["(", *self.normal.visit(expression), ")"]
        self.assertEqual(self.normal._format_function_args([expression]), expected)
        expected = self.normal.visit(expression)
        self.assertEqual(self.minified._format_function_args([expression]), expected)
        expression = parser._parse_exp()  # 1
        expected = ["(", *self.normal.visit(expression), ")"]
        self.assertEqual(self.minified._format_function_args([expression]), expected)
        expression = parser._parse_exp_list()  # "ghi", "jkl"
        expected = [
            "(",
            *self.normal.visit(expression[0]),
            Separators.Argument,
            *self.normal.visit(expression[1]),
            ")",
        ]
        self.assertEqual(self.normal._format_function_args(expression), expected)
        self.assertEqual(self.minified._format_function_args(expression), expected)

    def test_format_var(self):
        exp = Parser("a.b")._parse_var()
        expected = ["a", Separators.Dot, "b"]
        self.assertEqual(self.normal._format_var(exp), expected)
        exp = Parser("(a+b)")._parse_var()
        expected = ["(", "a", Separators.Space, "+", Separators.Space, "b", ")"]
        self.assertEqual(self.normal._format_var(exp), expected)

    def test_Assign(self):
        stmt = Parser("a = b")._parse_statement()
        expected = ["a", Separators.Space, "=", Separators.Space, "b"]
        self.assertEqual(self.normal.visit(stmt), expected)
        self.assertEqual(self.minified.visit(stmt), expected)

    def test_block(self):
        chunk = Parser("--comment\na=b c=d return a").parse_chunk()
        expected = [
            "-- comment",
            Separators.Newline,
            *self.normal.visit(chunk.statements[0]),
            Separators.Statement,
            *self.normal.visit(chunk.statements[1]),
            Separators.Statement,
            "return",
            Separators.Space,
            "a",
        ]
        self.assertEqual(self.normal.visit(chunk), expected)
        expected.pop(0)
        expected.pop(0)
        self.assertEqual(self.minified.visit(chunk), expected)
        chunk = Parser("a=b").parse_chunk()
        expected = [
            *self.normal.visit(chunk.statements[0]),
        ]
        self.assertEqual(self.normal.visit(chunk), expected)

    def test_Boolean(self):
        parser = Parser("true false")
        expr = parser._parse_exp()
        expected = ["true"]
        self.assertEqual(self.normal.visit(expr), expected)
        self.assertEqual(self.minified.visit(expr), expected)
        expr = parser._parse_exp()
        expected = ["false"]
        self.assertEqual(self.normal.visit(expr), expected)
        self.assertEqual(self.minified.visit(expr), expected)

    def test_Break(self):
        expr = Parser("break")._parse_break()
        expected = ["break"]
        self.assertEqual(self.normal.visit(expr), expected)

    def test_Goto(self):
        stmt = Parser("goto label")._parse_statement()
        expected = ["goto", Separators.Space, "label"]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_If(self):
        stmt = Parser("if 1 then a=b elseif 0 then c=d else e=f end")._parse_if()
        assert isinstance(stmt.false, If)
        assert isinstance(stmt.false.false, Block)
        expected = [
            "if",
            Separators.Space,
            "1",
            Separators.Space,
            "then",
            Separators.Block,
            *self.normal.visit(stmt.true)[2:-1],
            "elseif",
            Separators.Space,
            "0",
            Separators.Space,
            "then",
            Separators.Block,
            *self.normal.visit(stmt.false.true)[2:-1],
            "else",
            Separators.Block,
            *self.normal.visit(stmt.false.false)[2:-1],
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)
        stmt = Parser("if 1 then a=b end")._parse_if()
        expected = [
            "if",
            Separators.Space,
            "1",
            Separators.Space,
            "then",
            Separators.Block,
            *self.normal.visit(stmt.true)[2:-1],
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_Index(self):
        index = Parser("a[1]")._parse_var()
        expected = ["a", "[", "1", "]"]
        self.assertEqual(self.normal.visit(index), expected)

    def test_Label(self):
        stmt = Parser("::label::")._parse_statement()
        expected = ["::", "label", "::"]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_Name(self):
        name = Parser("abc")._parse_var()
        expected = ["abc"]
        self.assertEqual(self.normal.visit(name), expected)

    def test_Nil(self):
        exp = Parser("nil")._parse_exp()
        expected = ["nil"]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_Number(self):
        exp = Parser("1.23")._parse_exp()
        expected = ["1.23"]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_Repeat(self):
        stmt = Parser("repeat a=b until 1")._parse_repeat()
        expected = [
            "repeat",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "until",
            Separators.Space,
            "1",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_Semicolon(self):
        stmt = Parser(";")._parse_statement()
        self.assertEqual(self.normal.visit(stmt), [])
        self.assertEqual(self.test.visit(stmt), [";"])

    def test_String(self):
        parser = Parser(
            "'abc''def\"'[[ghi\njkl]][[mno'\"\"]][===[[[=[==[\n]===]'\\\\1\\x00'"
        )
        exp = parser._parse_exp()
        expected = ['"abc"']
        self.assertEqual(self.normal.visit(exp), expected)
        self.assertEqual(self.minified.visit(exp), expected)
        exp = parser._parse_exp()
        expected = ['"def\\""']
        self.assertEqual(self.normal.visit(exp), expected)
        expected = ["'def\"'"]
        self.assertEqual(self.minified.visit(exp), expected)
        exp = parser._parse_exp()
        expected = ['"ghi\\njkl"']
        self.assertEqual(self.normal.visit(exp), expected)
        self.assertEqual(self.minified.visit(exp), expected)
        exp = parser._parse_exp()
        expected = ['"mno\'\\"\\""']
        self.assertEqual(self.normal.visit(exp), expected)
        expected = ["'mno\\'\"\"'"]
        self.assertEqual(self.minified.visit(exp), expected)
        exp = parser._parse_exp()
        expected = ['"[[=[==[\\n"']
        self.assertEqual(self.normal.visit(exp), expected)
        self.assertEqual(self.minified.visit(exp), expected)
        exp = parser._parse_exp()
        expected = ['"\\\\1\\x00"']
        self.assertEqual(self.normal.visit(exp), expected)

    def test_Table(self):
        table = Parser("{a,b}")._parse_table_constructor()
        expected = ["{", "a", Separators.Argument, "b", "}"]
        self.assertEqual(self.normal.visit(table), expected)

    def test_Vararg(self):
        exp = Parser("...")._parse_exp()
        expected = ["..."]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_While(self):
        stmt = Parser("while i do a=b end")._parse_while()
        expected = [
            "while",
            Separators.Space,
            "i",
            Separators.Space,
            "do",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_BinOp(self):
        exp = Parser("a+b")._parse_exp()
        expected = ["a", Separators.Space, "+", Separators.Space, "b"]
        self.assertEqual(self.normal.visit(exp), expected)
        exp = Parser("(a+b)*(c+d)")._parse_exp()
        expected = [
            "(",
            "a",
            Separators.Space,
            "+",
            Separators.Space,
            "b",
            ")",
            Separators.Space,
            "*",
            Separators.Space,
            "(",
            "c",
            Separators.Space,
            "+",
            Separators.Space,
            "d",
            ")",
        ]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_FunctionCall(self):
        stmt = Parser("a(1)")._parse_statement()
        expected = ["a", "(", "1", ")"]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_FunctionDefinition(self):
        stmt = Parser("function a(b)a=b end")._parse_function()
        expected = [
            Separators.Newline,
            "function",
            Separators.Space,
            "a",
            "(",
            "b",
            ")",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "end",
            Separators.Block,
            Separators.Newline,
        ]
        self.assertEqual(self.normal.visit(stmt), expected)
        stmt = Parser("function a:b()a=b end")._parse_function()
        expected = [
            Separators.Newline,
            "function",
            Separators.Space,
            "a",
            ":",
            "b",
            "(",
            ")",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "end",
            Separators.Block,
            Separators.Newline,
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_IterativeFor(self):
        stmt = Parser("for a in b do a=b end")._parse_for()
        expected = [
            "for",
            Separators.Space,
            "a",
            Separators.Space,
            "in",
            Separators.Space,
            "b",
            Separators.Space,
            "do",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_LocalAssign(self):
        stmt = Parser("local a <const>, b = b")._parse_local()
        expected = [
            "local",
            Separators.Space,
            "a",
            Separators.Space,
            "<",
            "const",
            ">",
            Separators.Argument,
            "b",
            Separators.Space,
            "=",
            Separators.Space,
            "b",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)
        stmt = Parser("local a")._parse_local()
        expected = ["local", Separators.Space, "a"]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_MethodInvocation(self):
        stmt = Parser("a:b(1)")._parse_statement()
        expected = ["a", ":", "b", "(", "1", ")"]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_NamedIndex(self):
        exp = Parser("a.b")._parse_var()
        expected = ["a", Separators.Dot, "b"]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_NumericFor(self):
        stmt = Parser("for a=1,2 do a=b end")._parse_for()
        expected = [
            "for",
            Separators.Space,
            "a",
            Separators.Space,
            "=",
            Separators.Space,
            "1",
            Separators.Argument,
            "2",
            Separators.Space,
            "do",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)
        stmt = Parser("for a=1,2,3 do a=b end")._parse_for()
        expected = [
            "for",
            Separators.Space,
            "a",
            Separators.Space,
            "=",
            Separators.Space,
            "1",
            Separators.Argument,
            "2",
            Separators.Argument,
            "3",
            Separators.Space,
            "do",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_UnOp(self):
        exp = Parser("-1")._parse_exp()
        expected = ["-", "1"]
        self.assertEqual(self.normal.visit(exp), expected)
        exp = Parser("-(1+2)")._parse_exp()
        expected = ["-", "(", "1", Separators.Space, "+", Separators.Space, "2", ")"]
        self.assertEqual(self.normal.visit(exp), expected)
        exp = Parser("not 1")._parse_exp()
        expected = ["not", Separators.Space, "1"]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_ExpFunctionCall(self):
        exp = Parser("a(1)")._parse_exp()
        expected = ["a", "(", "1", ")"]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_ExpFunctionDefinition(self):
        exp = Parser("function (a,b)a=b end")._parse_exp()
        assert isinstance(exp, ExpFunctionDefinition)
        expected = [
            "function",
            Separators.Space,
            "(",
            "a",
            Separators.Argument,
            "b",
            ")",
            Separators.Block,
            *self.normal.visit(exp.body)[2:-1],
            "end",
        ]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_ExpMethodInvocation(self):
        exp = Parser("a:b(1)")._parse_exp()
        expected = ["a", ":", "b", "(", "1", ")"]
        self.assertEqual(self.normal.visit(exp), expected)

    def test_LocalFunctionDefinition(self):
        stmt = Parser("local function abc(a)a=b end")._parse_local()
        expected = [
            Separators.Newline,
            "local",
            Separators.Space,
            "function",
            Separators.Space,
            "abc",
            "(",
            "a",
            ")",
            Separators.Block,
            *self.normal.visit(stmt.body)[2:-1],
            "end",
            Separators.Statement,
            Separators.Newline,
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_ExplicitTableField(self):
        index = Parser("[1]=2")._parse_field()
        expected = ["[", "1", "]", Separators.Space, "=", Separators.Space, "2"]
        self.assertEqual(self.normal.visit(index), expected)

    def test_NamedTableField(self):
        index = Parser("a=2")._parse_field()
        expected = ["a", Separators.Space, "=", Separators.Space, "2"]
        self.assertEqual(self.normal.visit(index), expected)

    def test_NumberedTableField(self):
        index = Parser("a")._parse_field()
        expected = ["a"]
        self.assertEqual(self.normal.visit(index), expected)

    def test_return(self):
        stmt = Parser("do return end")._parse_statement()
        expected = [
            "do",
            Separators.Block,
            Separators.Indent,
            "return",
            Separators.Statement,
            Separators.DeIndent,
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)
        stmt = Parser("do end")._parse_statement()
        expected = [
            "do",
            Separators.Block,
            Separators.Indent,
            Separators.DeIndent,
            "end",
        ]
        self.assertEqual(self.normal.visit(stmt), expected)

    def test_non_commutative_brackets(self):
        expr = Parser("1 - (2 - 3)")._parse_exp()
        expected = [
            "1",
            Separators.Space,
            "-",
            Separators.Space,
            "(",
            "2",
            Separators.Space,
            "-",
            Separators.Space,
            "3",
            ")",
        ]
        self.assertEqual(self.normal.visit(expr), expected)
        expr = Parser("1 * (2 / 3)")._parse_exp()
        expected = [
            "1",
            Separators.Space,
            "*",
            Separators.Space,
            "(",
            "2",
            Separators.Space,
            "/",
            Separators.Space,
            "3",
            ")",
        ]
        self.assertEqual(self.normal.visit(expr), expected)


class TestTokenFormatters(unittest.TestCase):
    def test_sep_required(self):
        self.assertTrue(sep_required("a", "b"))
        self.assertTrue(sep_required("a", "1"))
        self.assertTrue(sep_required("1", "2"))
        self.assertTrue(sep_required("1", "b"))
        self.assertTrue(sep_required("-", "-"))
        self.assertFalse(sep_required("a", "("))
        self.assertFalse(sep_required("1", ","))

    def test_remove_separators(self):
        base = ["a", Separators.Space, "b", Separators.Space, "("]
        expected = ["a", Separators.Space, "b", "("]
        remove_separators(base)
        self.assertEqual(base, expected)
        base = [
            "a",
            Separators.Space,
            Separators.DeIndent,
            "b",
            Separators.Indent,
            Separators.Newline,
            "(",
        ]
        expected = [
            "a",
            Separators.Space,
            Separators.DeIndent,
            "b",
            Separators.Indent,
            Separators.Newline,
            "(",
        ]
        remove_separators(base)
        self.assertEqual(base, expected)

    def test_resolve_tokens(self):
        base = [
            Separators.Space,
            Separators.Indent,
            Separators.DeIndent,
            Separators.Statement,
            Separators.Newline,
            Separators.Argument,
            Separators.Dot,
            "a",
        ]
        expected = [
            " ",
            Separators.Indent,
            Separators.DeIndent,
            FormattingStyle.STATEMENT_SEPARATOR,
            FormattingStyle.STATEMENT_SEPARATOR,
            FormattingStyle.ARGUMENT_SEPARATOR,
            ".",
            "a",
        ]
        resolve_tokens(base, FormattingStyle)
        self.assertEqual(base, expected)
        base = [
            Separators.Space,
            Separators.Indent,
            Separators.DeIndent,
            Separators.Statement,
            Separators.Newline,
            Separators.Argument,
            Separators.Dot,
        ]
        expected = [
            " ",
            Separators.Indent,
            Separators.DeIndent,
            MinifiedStyle.STATEMENT_SEPARATOR,
            "\n",
            MinifiedStyle.ARGUMENT_SEPARATOR,
            ".",
        ]
        resolve_tokens(base, MinifiedStyle)
        self.assertEqual(base, expected)

    def test_indent(self):
        base = [
            "a\n",
            Separators.Indent,
            "b\n",
            Separators.Indent,
            "c\n",
            Separators.DeIndent,
            Separators.DeIndent,
            "a",
        ]
        expected = [
            "a\n",
            Separators.Indent,
            "\tb\n",
            Separators.Indent,
            "\t\tc\n",
            Separators.DeIndent,
            Separators.DeIndent,
            "a",
        ]
        indent(base, "\t")
        self.assertEqual(base, expected)

    def test_format(self):
        chunk = Parser("a=b if 1 then c=d else b=d end").parse_chunk()
        expected = "-- tumfl\na = b\nif 1 then\n\tc = d\nelse\n\tb = d\nend\n"
        self.assertEqual(format(chunk), expected)
        expected = "--tumfl\na=b;if 1 then;c=d;else;b=d;end"
        self.assertEqual(format(chunk, MinifiedStyle), expected)

    def test_string_ident(self):
        self.assertEqual(
            _string_ident("'abc def'", 0, TestStyle),
            ["'abc \\z", Separators.Newline, "def'"],
        )
        self.assertEqual(
            _string_ident("'ab def'", 0, TestStyle),
            ["'ab \\z", Separators.Newline, "def'"],
        )
        self.assertEqual(_string_ident("'          '", 0, TestStyle), ["'          '"])
        self.assertEqual(
            _string_ident("'      abc'", 0, TestStyle),
            ["'      \\z", Separators.Newline, "abc'"],
        )
        self.assertEqual(
            _string_ident("'abcdef'", 0, TestStyle),
            ["'abcd\\z", Separators.Newline, "ef'"],
        )

    def test_closure(self):
        chunk = Parser("(function()return 1 end)()").parse_chunk()
        self.assertEqual(format(chunk), "-- tumfl\n(function ()\n\treturn 1\nend)()\n")



from __future__ import annotations

from tumfl.Token import Token

from .Expression import Expression
from .Name import Name
from .Variable import Variable


class NamedIndex(Variable):
    """A named index, like a.b, c.e, NOT a["b"]"""

    def __init__(self, token: Token, lhs: Expression, variable_name: Name):
        super().__init__(token, "NamedIndex")
        self.lhs: Expression = lhs
        self.variable_name: Name = variable_name



from __future__ import annotations

from enum import Enum

from tumfl.Token import Token

from .Expression import Expression


class UnaryOperand(Enum):
    MINUS = "-"
    HASH = "#"
    BIT_XOR = "~"
    NOT = "not"

    def __repr__(self) -> str:
        return f"UnaryOperand.{self.name}"


class UnOp(Expression):
    """A unary operation, like -a, #c or not true"""

    def __init__(self, token: Token, op: UnaryOperand, right: Expression) -> None:
        super().__init__(token, "UnOp")
        self.op: UnaryOperand = op
        self.right: Expression = right

    @staticmethod
    def from_token(token: Token, right: Expression) -> UnOp:
        unop: UnaryOperand = UnaryOperand[token.type.name]
        assert unop
        return UnOp(token, unop, right)



import unittest

from tumfl.AST import *
from tumfl.AST.BaseFunctionDefinition import BaseFunctionDefinition
from tumfl.lexer import Lexer
from tumfl.parser import Parser, ParserError
from tumfl.Token import Token, TokenType


class TestParser(unittest.TestCase):
    @staticmethod
    def parse_number(to_parse: str) -> Number:
        lex = Lexer(to_parse)
        return Number.from_token(lex.get_next_token())

    @staticmethod
    def parse_name(name: str) -> Name:
        return Name.from_token(Token(TokenType.NAME, name, 0, 0))

    @staticmethod
    def parse_string(string: str) -> String:
        return String.from_token(Token(TokenType.STRING, string, 0, 0))

    @staticmethod
    def get_chunk(*statements: Statement) -> Chunk:
        return Chunk(statements[0].token, list(statements), None)

    def test_simple_exp(self):
        parser = Parser("1+2")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            self.parse_number("1"),
            self.parse_number("2"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("4+5*6")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            self.parse_number("4"),
            BinOp.from_token(
                Token(TokenType.MULT, "*", 0, 0),
                self.parse_number("5"),
                self.parse_number("6"),
            ),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("true and false or 1")
        expected_tree = BinOp.from_token(
            Token(TokenType.OR, "or", 0, 0),
            BinOp.from_token(
                Token(TokenType.AND, "and", 0, 0),
                Boolean.from_token(Token(TokenType.TRUE, "true", 0, 0)),
                Boolean.from_token(Token(TokenType.FALSE, "false", 0, 0)),
            ),
            self.parse_number("1"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("-nil")
        expected_tree = UnOp.from_token(
            Token(TokenType.MINUS, "-", 0, 0),
            Nil.from_token(Token(TokenType.NIL, "nil", 0, 0)),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("\n#'a'")
        expected_tree = UnOp.from_token(
            Token(TokenType.HASH, "#", 0, 0),
            self.parse_string("a"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("2^-2+s")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            BinOp.from_token(
                Token(TokenType.EXPONENT, "#", 0, 0),
                self.parse_number("2"),
                UnOp.from_token(
                    Token(TokenType.MINUS, "-", 0, 0), self.parse_number("2")
                ),
            ),
            self.parse_name("s"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_associativity(self):
        parser = Parser("1^2^3")
        expected_tree = BinOp.from_token(
            Token(TokenType.EXPONENT, "^", 0, 0),
            self.parse_number("1"),
            BinOp.from_token(
                Token(TokenType.EXPONENT, "^", 0, 0),
                self.parse_number("2"),
                self.parse_number("3"),
            ),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("1+2-3")
        expected_tree = BinOp.from_token(
            Token(TokenType.MINUS, "-", 0, 0),
            BinOp.from_token(
                Token(TokenType.PLUS, "+", 0, 0),
                self.parse_number("1"),
                self.parse_number("2"),
            ),
            self.parse_number("3"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_error(self):
        parser = Parser("1+")
        with self.assertRaises(ParserError):
            parser._parse_exp()

    def test_unary_expr(self):
        parser = Parser("1- -2")
        expected_tree = BinOp.from_token(
            Token(TokenType.MINUS, "-", 0, 0),
            self.parse_number("1"),
            UnOp.from_token(Token(TokenType.MINUS, "-", 0, 0), self.parse_number("2")),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_len_operator(self):
        parser = Parser("\n#a + 1")
        expected_tree = BinOp.from_token(
            Token(TokenType.PLUS, "+", 0, 0),
            UnOp.from_token(
                Token(TokenType.HASH, "#", 0, 0),
                self.parse_name("a"),
            ),
            self.parse_number("1"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_wrong_token(self):
        parser = Parser("a,b+")
        with self.assertRaises(ParserError):
            parser.parse_chunk()
        parser = Parser("a=1\nb,c+")
        with self.assertRaises(ParserError):
            parser.parse_chunk()

    def test_parse_var(self):
        parser = Parser('a["b"].d')
        expected_tree = NamedIndex(
            Token(TokenType.NAME, "d", 0, 0),
            Index(
                Token(TokenType.R_BRACKET, "[", 0, 0),
                self.parse_name("a"),
                self.parse_string("b"),
            ),
            self.parse_name("d"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_var(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("(2)[1]")
        expected_tree = Index(
            Token(TokenType.L_BRACKET, "[", 0, 0),
            self.parse_number("2"),
            self.parse_number("1"),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_var(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("(2)(1)")
        expected_tree = ExpFunctionCall(
            Token(TokenType.L_BRACKET, "[", 0, 0),
            self.parse_number("2"),
            [self.parse_number("1")],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_var(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("a:b()")
        expected_tree = ExpMethodInvocation(
            Token(TokenType.COLON, "[", 0, 0),
            self.parse_name("a"),
            self.parse_name("b"),
            [],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_var(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_table_constructor(self):
        parser = Parser("{}")
        expected_tree = Table(Token(TokenType.L_CURL, "{", 0, 0), [])
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("{a}")
        expected_tree = Table(
            Token(TokenType.L_CURL, "{", 0, 0),
            [
                NumberedTableField(
                    Token(TokenType.NAME, "a", 0, 0), self.parse_name("a")
                )
            ],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser('{["b"]=c}')
        expected_tree = Table(
            Token(TokenType.L_CURL, "{", 0, 0),
            [
                ExplicitTableField(
                    Token(TokenType.L_BRACKET, "[", 0, 0),
                    self.parse_string("b"),
                    self.parse_name("c"),
                )
            ],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("{a=b}")
        expected_tree = Table(
            Token(TokenType.L_CURL, "{", 0, 0),
            [
                NamedTableField(
                    Token(TokenType.NAME, "a", 0, 0),
                    self.parse_name("a"),
                    self.parse_name("b"),
                )
            ],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("{a, b,}")
        expected_tree = Table(
            Token(TokenType.L_CURL, "{", 0, 0),
            [
                NumberedTableField(
                    Token(TokenType.NAME, "a", 0, 0), self.parse_name("a")
                ),
                NumberedTableField(
                    Token(TokenType.NAME, "b", 0, 0), self.parse_name("b")
                ),
            ],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("{f'alo'..'xixi'}")
        expected_tree = Table(
            Token(TokenType.L_CURL, "{", 0, 0),
            [
                NumberedTableField(
                    Token(TokenType.NAME, "f", 0, 0),
                    BinOp.from_token(
                        Token(TokenType.CONCAT, "..", 0, 0),
                        ExpFunctionCall(
                            Token(TokenType.FUNCTION, "function", 0, 0),
                            self.parse_name("f"),
                            [self.parse_string("alo")],
                        ),
                        self.parse_string("xixi"),
                    ),
                )
            ],
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_table_constructor(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_args(self):
        parser = Parser("()")
        self.assertEqual(parser._parse_args(), [])
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser('(a, b, "d", 1)')
        expected_args = [
            self.parse_name("a"),
            self.parse_name("b"),
            self.parse_string("d"),
            self.parse_number("1"),
        ]
        self.assertEqual(parser._parse_args(), expected_args)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("(a,b")
        with self.assertRaises(ParserError):
            parser._parse_args()
        parser = Parser("(a,b,)")
        with self.assertRaises(ParserError):
            parser._parse_args()
        parser = Parser('"abc"')
        self.assertEqual(parser._parse_args(), [self.parse_string("abc")])
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser('{a,"b"}')
        table_result = parser._parse_table_constructor()
        parser = Parser('{a,"b"}')
        self.assertEqual(parser._parse_args(), [table_result])
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_names(self):
        parser = Parser("a")
        expected_names = [self.parse_name("a")]
        self.assertEqual(parser._parse_name_list(), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("a, b")
        expected_names = [self.parse_name("a"), self.parse_name("b")]
        self.assertEqual(parser._parse_name_list(), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("b")
        expected_names = [self.parse_name("a")]
        self.assertEqual(parser._parse_name_list(self.parse_name("a")), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.NAME, "b", 0, 0))
        parser = Parser(",b")
        expected_names = [self.parse_name("a"), self.parse_name("b")]
        self.assertEqual(parser._parse_name_list(self.parse_name("a")), expected_names)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("a,")
        with self.assertRaises(ParserError):
            parser._parse_name_list()
        parser = Parser("a, 1")
        with self.assertRaises(ParserError):
            parser._parse_name_list()

    def test_parse_vararg(self):
        parser = Parser("...")
        expected_tree = Vararg(Token(TokenType.ELLIPSIS, "...", 0, 0))
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_table_expr(self):
        parser = Parser("{a}")
        expected_tree = Parser("{a}")._parse_table_constructor()
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_funcbody(self):
        function_token = Token(TokenType.FUNCTION, "function", 0, 0)
        parser = Parser("()end")
        expected_tree = BaseFunctionDefinition(
            function_token, [], Block(Token(TokenType.END, "end", 0, 0), [], None)
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("(a,b)end")
        expected_tree = BaseFunctionDefinition(
            function_token,
            [self.parse_name("a"), self.parse_name("b")],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("(...)end")
        expected_tree = BaseFunctionDefinition(
            function_token,
            [Vararg.from_token(Token(TokenType.ELLIPSIS, "...", 0, 0))],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("(a,...)end")
        expected_tree = BaseFunctionDefinition(
            function_token,
            [
                self.parse_name("a"),
                Vararg.from_token(Token(TokenType.ELLIPSIS, "...", 0, 0)),
            ],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        parser._add_hint("function", "function")
        self.assertEqual(parser._parse_funcbody(function_token), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_func_expr(self):
        parser = Parser("function()end")
        expected_tree = ExpFunctionDefinition(
            Token(TokenType.FUNCTION, "function", 0, 0),
            [],
            Block(Token(TokenType.END, "end", 0, 0), [], None),
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser._parse_exp(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_assign(self):
        parser = Parser('a = "bcd"')
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.NAME, "a", 0, 0),
                [self.parse_name("a")],
                [self.parse_string("bcd")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser('a,b,c = "bcd"')
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.NAME, "a", 0, 0),
                [
                    self.parse_name("a"),
                    self.parse_name("b"),
                    self.parse_name("c"),
                ],
                [self.parse_string("bcd")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser('a = "bcd", "def"')
        expected_tree = self.get_chunk(
            Assign(
                Token(TokenType.NAME, "a", 0, 0),
                [self.parse_name("a")],
                [
                    self.parse_string("bcd"),
                    self.parse_string("def"),
                ],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_function_stmt(self):
        parser = Parser("a()")
        expected_tree = self.get_chunk(
            FunctionCall(Token(TokenType.NAME, "a", 0, 0), self.parse_name("a"), [])
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("a[1].b()")
        expected_tree = self.get_chunk(
            FunctionCall(
                Token(TokenType.NAME, "a", 0, 0),
                NamedIndex(
                    Token(TokenType.NAME, "b", 0, 0),
                    Index(
                        Token(TokenType.L_BRACKET, "[", 0, 0),
                        self.parse_name("a"),
                        self.parse_number("1"),
                    ),
                    self.parse_name("b"),
                ),
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_method_stmt(self):
        parser = Parser("a:b()")
        expected_tree = self.get_chunk(
            MethodInvocation(
                Token(TokenType.NAME, "a", 0, 0),
                self.parse_name("a"),
                self.parse_name("b"),
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_semicolon(self):
        parser = Parser(";")
        expected_tree = self.get_chunk(Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0)))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_local_assignment(self):
        parser = Parser("local a, b")
        expected_tree = self.get_chunk(
            LocalAssign(
                Token(TokenType.LOCAL, "local", 0, 0),
                [
                    AttributedName(self.parse_name("a")),
                    AttributedName(self.parse_name("b")),
                ],
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("local a <const>, b <final>")
        expected_tree = self.get_chunk(
            LocalAssign(
                Token(TokenType.LOCAL, "local", 0, 0),
                [
                    AttributedName(self.parse_name("a"), self.parse_name("const")),
                    AttributedName(self.parse_name("b"), self.parse_name("final")),
                ],
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("local a=d,e")
        expected_tree = self.get_chunk(
            LocalAssign(
                Token(TokenType.LOCAL, "local", 0, 0),
                [AttributedName(self.parse_name("a"))],
                [self.parse_name("d"), self.parse_name("e")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_local_function(self):
        parser = Parser("local function a ()end")
        expected_tree = self.get_chunk(
            LocalFunctionDefinition(
                Token(TokenType.LOCAL, "local", 0, 0),
                self.parse_name("a"),
                [],
                Block(Token(TokenType.END, "end", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_unknown_local(self):
        parser = Parser("local and")
        with self.assertRaises(ParserError):
            parser.parse_chunk()

    def test_parse_function(self):
        parser = Parser("function a ()end")
        expected_tree = self.get_chunk(
            FunctionDefinition(
                Token(TokenType.FUNCTION, "function", 0, 0),
                [self.parse_name("a")],
                None,
                [],
                Block(Token(TokenType.END, "end", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("function a.b:c ()end")
        expected_tree = self.get_chunk(
            FunctionDefinition(
                Token(TokenType.FUNCTION, "function", 0, 0),
                [self.parse_name("a"), self.parse_name("b")],
                self.parse_name("c"),
                [],
                Block(Token(TokenType.END, "end", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_iterative_for(self):
        parser = Parser("for a in 2 do end")
        expected_tree = self.get_chunk(
            IterativeFor(
                Token(TokenType.FOR, "for", 0, 0),
                [self.parse_name("a")],
                [self.parse_number("2")],
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("for a, b in 1, 2 do end")
        expected_tree = self.get_chunk(
            IterativeFor(
                Token(TokenType.FOR, "for", 0, 0),
                [self.parse_name("a"), self.parse_name("b")],
                [self.parse_number("1"), self.parse_number("2")],
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_numeric_for(self):
        parser = Parser("for a=1,b do end")
        expected_tree = self.get_chunk(
            NumericFor(
                Token(TokenType.FOR, "for", 0, 0),
                self.parse_name("a"),
                self.parse_number("1"),
                self.parse_name("b"),
                None,
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("for a=1,b,3 do end")
        expected_tree = self.get_chunk(
            NumericFor(
                Token(TokenType.FOR, "for", 0, 0),
                self.parse_name("a"),
                self.parse_number("1"),
                self.parse_name("b"),
                self.parse_number("3"),
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_unknown_for(self):
        parser = Parser("for name and")
        with self.assertRaises(ParserError):
            parser.parse_chunk()

    def test_parse_if(self):
        parser = Parser("if 1 then end")
        expected_tree = self.get_chunk(
            If(
                Token(TokenType.IF, "if", 0, 0),
                self.parse_number("1"),
                Block(Token(TokenType.THEN, "then", 0, 0), [], None),
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("if 1 then else end")
        expected_tree = self.get_chunk(
            If(
                Token(TokenType.IF, "if", 0, 0),
                self.parse_number("1"),
                Block(Token(TokenType.THEN, "then", 0, 0), [], None),
                Block(Token(TokenType.ELSE, "else", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("if 1 then elseif 2 then else end")
        expected_tree = self.get_chunk(
            If(
                Token(TokenType.IF, "if", 0, 0),
                self.parse_number("1"),
                Block(Token(TokenType.THEN, "then", 0, 0), [], None),
                If(
                    Token(TokenType.ELSEIF, "elseif", 0, 0),
                    self.parse_number("2"),
                    Block(Token(TokenType.ELSEIF, "elseif", 0, 0), [], None),
                    Block(Token(TokenType.ELSE, "else", 0, 0), [], None),
                ),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_repeat(self):
        parser = Parser("repeat until 1")
        expected_tree = self.get_chunk(
            Repeat(
                Token(TokenType.REPEAT, "repeat", 0, 0),
                self.parse_number("1"),
                Block(Token(TokenType.REPEAT, "repeat", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_while(self):
        parser = Parser("while 1 do end")
        expected_tree = self.get_chunk(
            While(
                Token(TokenType.WHILE, "while", 0, 0),
                self.parse_number("1"),
                Block(Token(TokenType.DO, "do", 0, 0), [], None),
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_label(self):
        parser = Parser("::label::")
        expected_tree = self.get_chunk(
            Label(Token(TokenType.LABEL_BORDER, "::", 0, 0), self.parse_name("label"))
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_goto(self):
        parser = Parser("goto label")
        expected_tree = self.get_chunk(
            Goto(Token(TokenType.GOTO, "goto", 0, 0), self.parse_name("label"))
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_break(self):
        parser = Parser("break")
        expected_tree = self.get_chunk(Break(Token(TokenType.BREAK, "break", 0, 0)))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_parse_block(self):
        parser = Parser("do end")
        expected_tree = self.get_chunk(Block(Token(TokenType.DO, "do", 0, 0), [], None))
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("do ; end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0))],
                None,
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("do ;return a; end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0))],
                [self.parse_name("a")],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("do ;return a end")
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("do ;return ; end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [Semicolon(Token(TokenType.SEMICOLON, ";", 0, 0))],
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))
        parser = Parser("do return end")
        expected_tree = self.get_chunk(
            Block(
                Token(TokenType.DO, "do", 0, 0),
                [],
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_unexpected_statement(self):
        parser = Parser("+")
        with self.assertRaises(ParserError):
            parser.parse_chunk()

    def test_weird_stuff(self):
        parser = Parser("assert()()")
        expected_tree = self.get_chunk(
            FunctionCall(
                Token(TokenType.NAME, "assert", 0, 0),
                ExpFunctionCall(
                    Token(TokenType.NAME, "assert", 0, 0), self.parse_name("assert"), []
                ),
                [],
            )
        )
        repr_test = repr(expected_tree)
        self.assertEqual(parser.parse_chunk(), expected_tree)
        self.assertEqual(len(parser.context_hints), 0)
        self.assertEqual(parser.current_token, Token(TokenType.EOF, "eof", 0, 0))

    def test_lonely_var(self):
        parser = Parser("a")
        with self.assertRaises(ParserError):
            parser.parse_chunk()

    def test_lonely_return(self):
        parser = Parser("do return end")
        expected_tree = self.get_chunk(Block(Token(TokenType.DO, "do", 0, 0), [], []))
        self.assertEqual(parser.parse_chunk(), expected_tree)
        parser = Parser("do end")
        expected_tree = self.get_chunk(Block(Token(TokenType.DO, "do", 0, 0), [], None))
        self.assertEqual(parser.parse_chunk(), expected_tree)


class EmptyTest(unittest.TestCase):
    def test_empty(self):
        parser = Parser("")
        parser.parse_chunk()



from __future__ import annotations

from tumfl.Token import Token

from .Expression import Expression
from .Name import Name


class ExpMethodInvocation(Expression):
    """A method invocation, may be a statement or an expression. Like f:b(a,b,c) or f:b()"""

    def __init__(
        self,
        token: Token,
        function: Expression,
        method: Name,
        arguments: list[Expression],
    ):
        super().__init__(token, "ExpMethodInvocation")
        self.function: Expression = function
        self.method: Name = method
        self.arguments: list[Expression] = arguments



from __future__ import annotations

import sys
from typing import Callable, NoReturn, Optional

from .AST import *
from .AST.BaseFunctionDefinition import BaseFunctionDefinition
from .error import ParserError
from .lexer import Lexer
from .Token import Token, TokenType


class Hint:
    def __init__(self, token: Token, where: str, what: str):
        self.token: Token = token
        self.where: str = where
        self.what: str = what

    def __str__(self) -> str:
        return f"<{self.where}> {self.what} ({self.token.line}:{self.token.column})"


class Parser:
    _BLOCK_END_TYPES = (
        TokenType.EOF,
        TokenType.END,
        TokenType.RETURN,
        TokenType.UNTIL,
        TokenType.ELSEIF,
        TokenType.ELSE,
    )

    def __init__(
        self, chunk: str, typed: bool = False, ignore_unicode_errors: bool = False
    ):
        self.chunk: str = chunk
        self.lexer: Lexer = Lexer(self.chunk, typed, ignore_unicode_errors)
        self.pos: int = 0
        self.current_token: Token = self.lexer.get_next_token()
        self.next_token: Token = self.lexer.get_next_token()
        self.context_hints: list[Hint] = []
        self.typed: bool = typed

    def _error(self, message: str, token: Token) -> NoReturn:
        """Throw an error, prints out a description, and finally throws a value error"""
        current_line: int = token.line
        current_column = token.column
        print(f"Error on line {current_line}:", file=sys.stderr)
        lines: list[str] = self.chunk.split("\n")
        if current_line > 1:
            print(lines[current_line - 2], file=sys.stderr)
        print(lines[current_line - 1], file=sys.stderr)
        print(" " * (current_column - 1) + "^", file=sys.stderr)
        print(message, file=sys.stderr)
        if self.context_hints:
            print(
                "hints:",
                " -> ".join(str(hint) for hint in self.context_hints),
                file=sys.stderr,
            )
        raise ParserError(message, self.context_hints, token)

    def _assert(self, token_type: TokenType) -> None:
        """Assert that the current token is of a certain kind"""
        if self.current_token.type != token_type:
            self._error("Unexpected token", self.current_token)

    def _eat_token(self, token_type: Optional[TokenType] = None) -> None:
        """Eats a specific token and throws an error, if the wrong token is found"""
        if token_type:
            self._assert(token_type)
        self.pos += 1
        self.current_token = self.next_token
        self.next_token = self.lexer.get_next_token()

    def _add_hint(self, where: str, what: str) -> None:
        """Add a context hint for error messages"""
        self.context_hints.append(Hint(self.current_token, where, what))

    def _remove_hint(self) -> None:
        """Remove a context hint for error messages"""
        self.context_hints.pop()

    def _switch_hint(self, what: str) -> None:
        self.context_hints[-1].what = what

    def parse_chunk(self) -> Chunk:
        """
        Parse a chunk

        chunk: block
        """
        block: Block = self._parse_block(self.current_token)
        return Chunk(block.token, block.statements, block.returns)

    def _parse_block(self, block_token: Token, expect_end: bool = False) -> Block:
        """
        Parse a block

        block: [DO | REPEAT | THEN | ELSE] {stat} [RETURN explist [SEMICOLON]] [END]
        """
        block: Block = Block(block_token, [], None)
        while self.current_token.type not in self._BLOCK_END_TYPES:
            block.statements.append(self._parse_statement())
        if self.current_token.type == TokenType.RETURN:
            self._eat_token(TokenType.RETURN)
            block.returns = []
            if (
                self.current_token.type
                not in (TokenType.SEMICOLON,) + self._BLOCK_END_TYPES
            ):
                block.returns = self._parse_exp_list()
            if self.current_token.type == TokenType.SEMICOLON:  # type: ignore
                self._eat_token()
        if expect_end:
            self._eat_token(TokenType.END)
        return block

    def _parse_statement(self) -> Statement:
        """Parse a statement"""
        match self.current_token.type:
            case TokenType.SEMICOLON:
                return self._parse_semi()
            case TokenType.BREAK:
                return self._parse_break()
            case TokenType.GOTO:
                return self._parse_goto()
            case TokenType.LABEL_BORDER:
                return self._parse_label()
            case TokenType.DO:
                block_token: Token = self.current_token
                self._eat_token()
                return self._parse_block(block_token, True)
            case TokenType.WHILE:
                return self._parse_while()
            case TokenType.REPEAT:
                return self._parse_repeat()
            case TokenType.IF:
                return self._parse_if()
            case TokenType.FOR:
                return self._parse_for()
            case TokenType.FUNCTION:
                return self._parse_function()
            case TokenType.LOCAL:
                return self._parse_local()
            case TokenType.L_PAREN | TokenType.NAME:
                return self._parse_var_stmt()
        self._error("Unexpected statement", self.current_token)

    def _parse_break(self) -> Break:
        """Parse a break statement"""
        break_token: Token = self.current_token
        self._eat_token(TokenType.BREAK)
        return Break(break_token)

    def __eat_name(self) -> Name:
        """Eat a name token and return the string value"""
        name_token: Token = self.current_token
        self._eat_token(TokenType.NAME)
        return Name.from_token(name_token)

    def _parse_goto(self) -> Goto:
        """
        Parse a goto statement

        goto_stmt: GOTO Name
        """
        goto_token: Token = self.current_token
        self._add_hint("goto", "name")
        self._eat_token(TokenType.GOTO)
        name: Name = self.__eat_name()
        self._remove_hint()
        return Goto(goto_token, name)

    def _parse_label(self) -> Label:
        """
        Parse a label statement

        label_stmt: LABEL_BORDER Name LABEL_BORDER
        """
        opening_token: Token = self.current_token
        self._add_hint("label", "name")
        self._eat_token(TokenType.LABEL_BORDER)
        name: Name = self.__eat_name()
        self._switch_hint("end")
        self._eat_token(TokenType.LABEL_BORDER)
        self._remove_hint()
        return Label(opening_token, name)

    def _parse_while(self) -> While:
        """
        Parse a while loop

        while_stmt: WHILE exp DO block END
        """
        while_token: Token = self.current_token
        self._add_hint("while", "condition")
        self._eat_token(TokenType.WHILE)
        condition: Expression = self._parse_exp()
        self._switch_hint("block")
        block_token: Token = self.current_token
        self._eat_token(TokenType.DO)
        body: Block = self._parse_block(block_token, True)
        body.comment.extend(while_token.comment)
        self._remove_hint()
        return While(while_token, condition, body)

    def _parse_repeat(self) -> Repeat:
        """
        Parse a repeat until loop

        repeat_stmt: REPEAT block UNTIL exp
        """
        repeat_token: Token = self.current_token
        self._add_hint("repeat", "block")
        block_token: Token = self.current_token
        self._eat_token(TokenType.REPEAT)
        body: Block = self._parse_block(block_token)
        body.comment.extend(repeat_token.comment)
        self._switch_hint("condition")
        self._eat_token(TokenType.UNTIL)
        condition: Expression = self._parse_exp()
        self._remove_hint()
        return Repeat(repeat_token, condition, body)

    def _parse_if(self) -> If:
        """
        Parse an if statement

        if_stmt: IF exp THEN block {ELSEIF exp THEN block} [ELSE block] end
        """
        if_token: Token = self.current_token
        self._add_hint("if", "if condition")
        self._eat_token(TokenType.IF)
        if_cond: Expression = self._parse_exp()
        self._switch_hint("if block")
        block_token: Token = self.current_token
        self._eat_token(TokenType.THEN)
        if_block: Block = self._parse_block(block_token)
        if_block.comment.extend(if_token.comment)
        elseif_tokens: list[Token] = []
        elseif_conditions: list[Expression] = []
        elseif_blocks: list[Block] = []
        while self.current_token.type == TokenType.ELSEIF:
            self._switch_hint("elseif condition")
            elseif_tokens.append(self.current_token)
            self._eat_token()
            elseif_conditions.append(self._parse_exp())
            self._switch_hint("elseif block")
            block_token = self.current_token
            self._eat_token(TokenType.THEN)
            elseif_blocks.append(self._parse_block(block_token))
        else_block: Optional[Block] = None
        if self.current_token.type == TokenType.ELSE:
            self._switch_hint("else block")
            block_token = self.current_token
            self._eat_token()
            else_block = self._parse_block(block_token)
        self._eat_token(TokenType.END)
        self._remove_hint()
        assert len(elseif_conditions) == len(elseif_blocks) and len(
            elseif_blocks
        ) == len(elseif_tokens)
        current_if: Optional[If | Block] = else_block
        for i in range(len(elseif_conditions) - 1, -1, -1):
            elseif_blocks[i].comment.extend(if_token.comment)
            current_if = If(
                elseif_tokens[i], elseif_conditions[i], elseif_blocks[i], current_if
            )
        return If(if_token, if_cond, if_block, current_if)

    def _parse_for(self) -> IterativeFor | NumericFor:
        """
        Parses either an iterative or numeric for, depending on how it looks like

        for_stmt: num_for | it_for
        """
        for_token: Token = self.current_token
        self._add_hint("for", "name")
        self._eat_token(TokenType.FOR)
        first_name: Name = self.__eat_name()
        self._remove_hint()
        if self.current_token.type == TokenType.ASSIGN:
            return self._parse_numeric_for(for_token, first_name)
        if self.current_token.type in (TokenType.COMMA, TokenType.IN):
            return self._parse_iterative_for(for_token, first_name)
        self._error("unexpected for condition", self.current_token)

    def _parse_numeric_for(self, for_token: Token, name: Name) -> NumericFor:
        """
        Parse a numeric for

        num_for: FOR Name ASSIGN exp COMMA exp [COMMA exp] DO block END
        """
        self._add_hint("numeric for", "start expression")
        self._eat_token(TokenType.ASSIGN)
        start: Expression = self._parse_exp()
        self._switch_hint("stop expression")
        self._eat_token(TokenType.COMMA)
        stop: Expression = self._parse_exp()
        step: Optional[Expression] = None
        if self.current_token.type == TokenType.COMMA:
            self._switch_hint("step expression")
            self._eat_token()
            step = self._parse_exp()
        self._switch_hint("block")
        block_token: Token = self.current_token
        self._eat_token(TokenType.DO)
        body: Block = self._parse_block(block_token, True)
        body.comment.extend(for_token.comment)
        self._remove_hint()
        return NumericFor(for_token, name, start, stop, step, body)

    def _parse_iterative_for(self, for_token: Token, name: Name) -> IterativeFor:
        """
        Parse an iterative for

        it_for: FOR namelist IN explist DO block END
        """
        self._add_hint("iterative for", "name list")
        names: list[Name] = self._parse_name_list(name)
        self._switch_hint("expression list")
        self._eat_token(TokenType.IN)
        expressions: list[Expression] = self._parse_exp_list()
        self._switch_hint("block")
        block_token: Token = self.current_token
        self._eat_token(TokenType.DO)
        body: Block = self._parse_block(block_token, True)
        body.comment.extend(for_token.comment)
        self._remove_hint()
        return IterativeFor(for_token, names, expressions, body)

    def _parse_function(self) -> FunctionDefinition:
        """
        Parse a non-local function statement

        funct_stmt: FUNCTION Name {DOT Name} [COLON Name] funcbody
        """
        function_token: Token = self.current_token
        self._add_hint("function", "name")
        self._eat_token(TokenType.FUNCTION)
        names: list[Name] = [self.__eat_name()]
        while self.current_token.type == TokenType.DOT:
            self._eat_token()
            names.append(self.__eat_name())
        method: Optional[Name] = None
        if self.current_token.type == TokenType.COLON:
            self._switch_hint("method name")
            self._eat_token()
            method = self.__eat_name()
        base_function: BaseFunctionDefinition = self._parse_funcbody(function_token)
        return FunctionDefinition.from_base_definition(base_function, names, method)

    def _parse_funcbody(self, function_token: Token) -> BaseFunctionDefinition:
        """
        Parse a function body (starting from parameters till end)

        funcbody: L_PAREN [namelist [COMMA ELLIPSIS] | ELLIPSIS] R_PAREN block END
        """
        self._switch_hint("parameters")
        self._eat_token(TokenType.L_PAREN)
        parameters: list[Name | Vararg] = []
        if self.current_token.type == TokenType.NAME:
            parameters.extend(self._parse_name_list(leave_vararg=True))
            if self.current_token.type == TokenType.ELLIPSIS:  # type: ignore
                self._switch_hint("varargs")
                parameters.append(Vararg.from_token(self.current_token))
                self._eat_token()
        elif self.current_token.type == TokenType.ELLIPSIS:
            self._switch_hint("varargs")
            parameters.append(Vararg.from_token(self.current_token))
            self._eat_token()
        block_token: Token = self.current_token
        self._switch_hint("body")
        self._eat_token(TokenType.R_PAREN)
        body: Block = self._parse_block(block_token, True)
        body.comment.extend(function_token.comment)
        self._remove_hint()
        return BaseFunctionDefinition(function_token, parameters, body)

    def _parse_local(self) -> LocalFunctionDefinition | LocalAssign:
        """
        Parse a local function definition or local assign

        local_stmt: local_funct_stmt | local_assign_stmt
        """
        local_token: Token = self.current_token
        self._eat_token(TokenType.LOCAL)
        if self.current_token.type == TokenType.FUNCTION:
            return self._parse_local_function(local_token)
        if self.current_token.type == TokenType.NAME:
            return self._parse_local_assignment(local_token)
        self._error("Unexpected symbol after local", self.current_token)

    def _parse_local_function(self, local_token: Token) -> LocalFunctionDefinition:
        """
        Parse a local function definition

        local_funct_stmt: LOCAL FUNCTION Name funcbody
        """
        self._add_hint("local function", "name")
        function_token: Token = self.current_token
        function_token.comment.extend(local_token.comment)
        self._eat_token(TokenType.FUNCTION)
        name: Name = self.__eat_name()
        base_function: BaseFunctionDefinition = self._parse_funcbody(function_token)
        return LocalFunctionDefinition.from_base_definition(base_function, name)

    def _parse_local_assignment(self, local_token: Token) -> LocalAssign:
        """
        Parse a local assign

        local_assign_stmt: LOCAL Name [LESS_THAN Name GREATER_THAN] {COMMA Name [LESS_THAN Name GREATER_THAN]} [ASSIGN explist]
        """
        self._add_hint("local assign", "names")
        self._assert(TokenType.NAME)
        names: list[AttributedName] = []
        while True:
            self._switch_hint("name")
            name: Name = self.__eat_name()
            attribute: Optional[Name] = None
            if self.current_token.type == TokenType.LESS_THAN:
                self._switch_hint("name attribute")
                self._eat_token()
                attribute = self.__eat_name()
                self._eat_token(TokenType.GREATER_THAN)
            names.append(AttributedName(name, attribute))
            if self.current_token.type == TokenType.COMMA:
                self._eat_token()
            else:
                break
        expressions: Optional[list[Expression]] = None
        if self.current_token.type == TokenType.ASSIGN:
            self._switch_hint("expressions")
            self._eat_token()
            expressions = self._parse_exp_list()
        self._remove_hint()
        return LocalAssign(local_token, names, expressions)

    def _parse_exp_list(self) -> list[Expression]:
        """
        Parse a list of expressions

        explist: exp {COMMA exp}
        """
        expressions: list[Expression] = []
        while True:
            expressions.append(self._parse_exp())
            if self.current_token.type == TokenType.COMMA:
                self._eat_token()
            else:
                break
        return expressions

    def _parse_semi(self) -> Semicolon:
        """
        Parse a semicolon

        semi_stmt: SEMICOLON
        """
        semicolon_token: Token = self.current_token
        self._eat_token(TokenType.SEMICOLON)
        return Semicolon(semicolon_token)

    def _parse_var_stmt(self) -> FunctionCall | MethodInvocation | Assign:
        """
        Parse a variable assignment or a function

        var_funct: assign_stmt | funct_call_stmt
        """
        first_token: Token = self.current_token
        first_var: Expression = self._parse_var()
        if self.current_token.type in (TokenType.COMMA, TokenType.ASSIGN):
            return self._parse_assignment(first_token, first_var)
        if isinstance(first_var, ExpFunctionCall):
            return FunctionCall(
                first_var.token, first_var.function, first_var.arguments
            )
        if isinstance(first_var, ExpMethodInvocation):
            return MethodInvocation(
                first_var.token,
                first_var.function,
                first_var.method,
                first_var.arguments,
            )
        if isinstance(first_var, Name):
            self._error("Unexpected lonely name", self.current_token)
        assert False, "unreachable line"

    def _parse_assignment(self, first_token: Token, first_var: Expression) -> Assign:
        """
        Parse an assignment

        assign_stmt: var {COMMA var} ASSIGN explist
        """
        variables: list[Expression] = [first_var]
        self._add_hint("assignment", "variables")
        while self.current_token.type == TokenType.COMMA:
            self._eat_token()
            variables.append(self._parse_var())
        self._switch_hint("expressions")
        self._eat_token(TokenType.ASSIGN)
        expressions: list[Expression] = self._parse_exp_list()
        self._remove_hint()
        return Assign(first_token, variables, expressions)

    def __parse_left_associative_binop(
        self, types: tuple[TokenType, ...], base: Callable[[], Expression]
    ) -> Expression:
        """Parse a generic left associative binop"""
        node: Expression = base()
        while self.current_token.type in types:
            token: Token = self.current_token
            self._eat_token()
            node = BinOp.from_token(token, node, base())
        return node

    def __parse_right_associative_binop(
        self,
        type: TokenType,
        base: Callable[[], Expression],
        operand: Callable[[], Expression],
    ) -> Expression:
        """Parse a generic right associative binop"""
        nodes: list[Expression] = [base()]
        tokens: list[Token] = []
        while self.current_token.type == type:
            tokens.append(self.current_token)
            self._eat_token()
            nodes.append(operand())
        node: Expression = nodes[-1]
        for i in range(len(nodes) - 2, -1, -1):
            node = BinOp.from_token(tokens[i], nodes[i], node)
        return node

    def _parse_exp(self) -> Expression:
        """
        Parse an expression

        exp: and_exp {OR and_exp}
        """
        return self.__parse_left_associative_binop((TokenType.OR,), self._parse_and_exp)

    def _parse_and_exp(self) -> Expression:
        """
        Parse an and expression

        and_exp: comp_exp {OR comp_exp}
        """
        return self.__parse_left_associative_binop(
            (TokenType.AND,), self._parse_comp_exp
        )

    def _parse_comp_exp(self) -> Expression:
        """
        Parse a comparative expression

        comp_exp: bit_or_exp
                  {(LESS_THAN | GREATER_THAN | LESS_EQUALS | GREATER_EQUALS | NOT_EQUALS | EQUALS) bit_or_exp}
        """
        return self.__parse_left_associative_binop(
            (
                TokenType.LESS_THAN,
                TokenType.GREATER_THAN,
                TokenType.LESS_EQUALS,
                TokenType.GREATER_EQUALS,
                TokenType.NOT_EQUALS,
                TokenType.EQUALS,
            ),
            self._parse_bit_or_exp,
        )

    def _parse_bit_or_exp(self) -> Expression:
        """
        Parse a bitwise or expression

        bit_or_exp: bit_xor_exp {BIT_OR bit_xor_exp}
        """
        return self.__parse_left_associative_binop(
            (TokenType.BIT_OR,), self._parse_bit_xor_exp
        )

    def _parse_bit_xor_exp(self) -> Expression:
        """
        Parse a bitwise and expression

        bit_xor_exp: bit_and_expr {BIT_XOR bit_and_expr}
        """
        return self.__parse_left_associative_binop(
            (TokenType.BIT_XOR,), self._parse_bit_and_exp
        )

    def _parse_bit_and_exp(self) -> Expression:
        """
        Parse a bitwise and expression

        bit_and_exp: bit_shift_exp {BIT_AND bit_shift_exp}
        """
        return self.__parse_left_associative_binop(
            (TokenType.BIT_AND,), self._parse_bit_shift_exp
        )

    def _parse_bit_shift_exp(self) -> Expression:
        """
        Parse a bit shift expression

        bit_shift_exp: concat_exp {(BIT_SHIFT_LEFT | BIT_SHIFT_RIGHT) concat_exp}
        """
        return self.__parse_left_associative_binop(
            (TokenType.BIT_SHIFT_LEFT, TokenType.BIT_SHIFT_RIGHT),
            self._parse_concat_exp,
        )

    def _parse_concat_exp(self) -> Expression:
        """
        Parse a concat expression, right associative

        concat_exp: add_exp {CONCAT add_exp}
        """
        return self.__parse_right_associative_binop(
            TokenType.CONCAT, self._parse_add_exp, self._parse_concat_exp
        )

    def _parse_add_exp(self) -> Expression:
        """
        Parse a addition or subtraction expression

        add_exp: mul_exp {(PLUS | MINUS) mul_exp}
        """
        return self.__parse_left_associative_binop(
            (TokenType.PLUS, TokenType.MINUS), self._parse_mul_exp
        )

    def _parse_mul_exp(self) -> Expression:
        """
        Parse a multiplication or division expression

        mul_exp: un_exp {(MULT | DIVIDE | INTEGER_DIVISION | MODULO) un_exp}
        """
        return self.__parse_left_associative_binop(
            (
                TokenType.MULT,
                TokenType.DIVIDE,
                TokenType.INTEGER_DIVISION,
                TokenType.MODULO,
            ),
            self._parse_un_exp,
        )

    def _parse_un_exp(self) -> Expression:
        """
        Parse an unary expression

        un_exp: pow_exp | NOT un_exp | HASH un_exp | MINUS un_exp | BIT_XOR un_exp
        """
        if self.current_token.type in (
            TokenType.NOT,
            TokenType.HASH,
            TokenType.MINUS,
            TokenType.BIT_XOR,
        ):
            token = self.current_token
            self._eat_token()
            return UnOp.from_token(token, self._parse_un_exp())
        return self._parse_pow_exp()

    def _parse_pow_exp(self) -> Expression:
        """
        Parse a exponentiation expression, right associative

        pow_exp: atom {EXPONENT atom}
        """
        return self.__parse_right_associative_binop(
            TokenType.EXPONENT, self._parse_atom, self._parse_un_exp
        )

    def _parse_atom(self) -> Expression:
        """
        Parse an atom

        atom: NIL | TRUE | FALSE | NUMBER | STRING | ELLIPSIS | FUNCTION func_def | tableconstructor | var
        """
        token: Token = self.current_token
        if token.type == TokenType.NIL:
            self._eat_token()
            return Nil.from_token(token)
        if token.type == TokenType.TRUE:
            self._eat_token()
            return Boolean.from_token(token)
        if token.type == TokenType.FALSE:
            self._eat_token()
            return Boolean.from_token(token)
        if token.type == TokenType.NUMBER:
            self._eat_token()
            return Number.from_token(token)
        if token.type == TokenType.STRING:
            self._eat_token()
            return String.from_token(token)
        if token.type == TokenType.ELLIPSIS:
            self._eat_token()
            return Vararg.from_token(token)
        if token.type == TokenType.FUNCTION:
            function_token: Token = self.current_token
            self._add_hint("function expression", "definition")
            self._eat_token()
            base_function: BaseFunctionDefinition = self._parse_funcbody(function_token)
            return ExpFunctionDefinition.from_base_definition(base_function)
        if token.type == TokenType.L_CURL:
            return self._parse_table_constructor()
        if token.type in (TokenType.L_PAREN, TokenType.NAME):
            return self._parse_var()
        self._error("Unexpected expression", self.current_token)

    def _parse_name_list(
        self, first_name: Optional[Name] = None, leave_vararg: bool = False
    ) -> list[Name]:
        """
        Parse a list of names

        name_list: Name {COMMA Name}
        """
        names: list[Name] = [first_name] if first_name else []
        if first_name and self.current_token.type == TokenType.COMMA:
            self._eat_token()
        elif first_name:
            return names
        while True:
            if self.current_token.type == TokenType.ELLIPSIS and leave_vararg:
                return names
            names.append(self.__eat_name())
            if self.current_token.type == TokenType.COMMA:
                self._eat_token()
            else:
                break
        return names

    def _parse_var(self) -> Expression:
        """
        Parse a variable. Might be either a Variable, function call, method invocation, index, named index,
        or any other expression

        var: Name | L_PAREN exp R_PAREN
        """
        var: Expression
        if self.current_token.type == TokenType.NAME:
            self._add_hint("named var", "name")
            var = self.__eat_name()
        elif self.current_token.type == TokenType.L_PAREN:
            self._add_hint("expression var", "expression")
            self._eat_token()
            var = self._parse_exp()
            self._eat_token(TokenType.R_PAREN)
        else:
            assert False
        var = self._parse_or_ignore_var_terminal(var)
        self._remove_hint()
        return var

    def _parse_or_ignore_var_terminal(self, base_var: Expression) -> Expression:
        """
        Parse a var terminal if necessary
        """
        if self.current_token.type in (
            TokenType.L_BRACKET,
            TokenType.DOT,
            TokenType.L_PAREN,
            TokenType.L_CURL,
            TokenType.COLON,
            TokenType.STRING,
        ):
            base_var = self._parse_var_terminal(base_var)
        return base_var

    def _parse_var_terminal(self, base_var: Expression) -> Expression:
        """
        Parse a variable terminal,

        var_terminal: var | var_terminal L_BRACKET exp R_BRACKET | var_terminal DOT Name | var_terminal args |
            var_terminal COLON Name args
        """
        token: Token = self.current_token
        var: Expression
        name: Name
        match token.type:
            case TokenType.L_PAREN | TokenType.L_CURL | TokenType.STRING:
                self._add_hint("function", "arguments")
                var = ExpFunctionCall(token, base_var, self._parse_args())
            case TokenType.COLON:
                self._add_hint("invocation", "name")
                self._eat_token()
                name = self.__eat_name()
                args: list[Expression] = self._parse_args()
                var = ExpMethodInvocation(token, base_var, name, args)
            case TokenType.L_BRACKET:
                self._add_hint("index", "expression")
                self._eat_token()
                expr: Expression = self._parse_exp()
                self._eat_token(TokenType.R_BRACKET)
                var = Index(token, base_var, expr)
            case TokenType.DOT:
                self._add_hint("index", "name")
                self._eat_token()
                name = self.__eat_name()
                var = NamedIndex(token, base_var, name)
            case _:  # pragma: no cover
                assert False, "unreachable line"
        var = self._parse_or_ignore_var_terminal(var)
        self._remove_hint()
        return var

    def _parse_table_constructor(self) -> Table:
        """
        Parse a table constructor

        tableconstructor: L_CURL [field {(COMMA | SEMICOLON) field} (COMMA | SEMICOLON)] R_CURL
        """
        table: Table = Table(self.current_token, [])
        self._add_hint("table constructor", "fields")
        self._eat_token(TokenType.L_CURL)
        while self.current_token.type not in (TokenType.R_CURL, TokenType.EOF):
            table.fields.append(self._parse_field())
            if self.current_token.type in (TokenType.COMMA, TokenType.SEMICOLON):
                self._eat_token()
        self._eat_token(TokenType.R_CURL)
        self._remove_hint()
        return table

    def _parse_field(self) -> TableField:
        """
        Parse a table field

        field: L_BRACKET exp R_BRACKET ASSIGN exp | Name ASSIGN exp | exp
        """
        token: Token = self.current_token
        value: Expression
        if self.current_token.type == TokenType.L_BRACKET:
            self._add_hint("explicit table field", "key expression")
            self._eat_token()
            at: Expression = self._parse_exp()
            self._eat_token(TokenType.R_BRACKET)
            self._switch_hint("value expression")
            self._eat_token(TokenType.ASSIGN)
            value = self._parse_exp()
            self._remove_hint()
            return ExplicitTableField(token, at, value)
        if (
            self.current_token.type == TokenType.NAME
            and self.next_token.type == TokenType.ASSIGN
        ):
            self._add_hint("named table field", "name")
            name: Name = self.__eat_name()
            self._switch_hint("value expression")
            self._eat_token(TokenType.ASSIGN)
            value = self._parse_exp()
            self._remove_hint()
            return NamedTableField(token, name, value)
        self._add_hint("numbered table field", "expression")
        value = self._parse_exp()
        self._remove_hint()
        return NumberedTableField(token, value)

    def _parse_args(self) -> list[Expression]:
        """
        Parse an argument array for a function call

        args: L_PAREN [explist] R_PAREN | tableconstructor | LiteralString
        """
        self._add_hint("function call", "arguments")
        expressions: list[Expression] = []
        if self.current_token.type == TokenType.L_PAREN:
            self._eat_token()
            if self.current_token.type != TokenType.R_PAREN:  # type: ignore
                expressions = self._parse_exp_list()
            self._eat_token(TokenType.R_PAREN)
        elif self.current_token.type == TokenType.L_CURL:
            expressions.append(self._parse_table_constructor())
        elif self.current_token.type == TokenType.STRING:
            expressions.append(String.from_token(self.current_token))
            self._eat_token()
        else:
            assert False, "unreachable line"
        self._remove_hint()
        return expressions


def parse(chunk: str) -> Chunk:
    ast: Chunk = Parser(chunk).parse_chunk()
    ast.parent(None)
    return ast



from __future__ import annotations

from tumfl.AST.Expression.Expression import Expression
from tumfl.AST.Expression.Name import Name
from tumfl.Token import Token

from .Block import Block
from .Statement import Statement


class IterativeFor(Statement):
    """The iterative for, i.e. using pairs() or ipairs()"""

    def __init__(
        self,
        token: Token,
        namelist: list[Name],
        explist: list[Expression],
        body: Block,
    ):
        super().__init__(token, "IterativeFor")
        self.namelist: list[Name] = namelist
        self.explist: list[Expression] = explist
        self.body: Block = body



from __future__ import annotations

from tumfl.Token import Token, TokenType

from .Expression import Expression


class String(Expression):
    """A lua string, like "abc" or [[abc]]"""

    def __init__(self, token: Token, value: str) -> None:
        super().__init__(token, "String")
        self.value: str = value

    @staticmethod
    def from_token(token: Token) -> String:
        assert token.type == TokenType.STRING
        value = token.value
        assert isinstance(value, str)
        return String(token, value)



import unittest
from pathlib import Path

from tumfl.dependency_resolver import (
    Chunk,
    FunctionCall,
    InvalidDependencyError,
    Semicolon,
    resolve_recursive,
)


class TestDependencyResolver(unittest.TestCase):
    def setUp(self) -> None:
        self.base_path: Path = Path(__file__).parent / "test_files"
        self.main_script: Path = self.base_path / "main.lua"
        assert self.main_script.is_file()

    def test_resolver(self) -> None:
        ast = resolve_recursive(self.main_script, [self.base_path])
        assert isinstance(ast, Chunk)
        self.assertIsInstance(ast.statements[0], Chunk)
        self.assertIsInstance(ast.statements[1], Chunk)
        self.assertIsInstance(ast.statements[2], Semicolon)
        self.assertIsInstance(ast.statements[3], FunctionCall)

    def test_errors(self) -> None:
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(self.base_path / "empty_path.lua", [self.base_path])
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(self.base_path / "nonexistent_path.lua", [self.base_path])
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(self.base_path / "dual_path.lua", [self.base_path])
        with self.assertRaises(InvalidDependencyError):
            resolve_recursive(self.base_path / "number_path.lua", [self.base_path])



from .Assign import Assign
from .Block import Block
from .Break import Break
from .Chunk import Chunk
from .FunctionCall import FunctionCall
from .FunctionDefinition import FunctionDefinition
from .Goto import Goto
from .If import If
from .IterativeFor import IterativeFor
from .Label import Label
from .LocalAssign import AttributedName, LocalAssign
from .LocalFunctionDefinition import LocalFunctionDefinition
from .MethodInvocation import MethodInvocation
from .NumericFor import NumericFor
from .Repeat import Repeat
from .Semicolon import Semicolon
from .Statement import Statement
from .While import While



from __future__ import annotations

from tumfl.Token import Token

from .Statement import Statement


class Break(Statement):
    """The break statement to exit loops"""

    def __init__(self, token: Token):
        super().__init__(token, "Break")



from __future__ import annotations

from abc import ABC

from tumfl.AST.ASTNode import ASTNode
from tumfl.Token import Token

from .Expression import Expression
from .Name import Name


class TableField(ASTNode, ABC):
    """Base class for all table fields"""


class ExplicitTableField(TableField):
    """Table field of the form `[ expr ] = expr`"""

    def __init__(self, token: Token, at: Expression, value: Expression):
        super().__init__(token, "ExplicitTableField")
        self.at: Expression = at
        self.value: Expression = value


class NamedTableField(TableField):
    """Table field of the form `Name = expr`"""

    def __init__(self, token: Token, field_name: Name, value: Expression):
        super().__init__(token, "NamedTableField")
        self.field_name: Name = field_name
        self.value: Expression = value


class NumberedTableField(TableField):
    """Table field of the form `expr`"""

    def __init__(self, token: Token, value: Expression):
        super().__init__(token, "NumberedTableField")
        self.value: Expression = value
